import copy
import functools
from typing import Sequence

from pyos import db
from pyos import fs
from pyos import os
from pyos import pathlib

__all__ = ('parse_args',)


@functools.singledispatch
def parse_arg(arg) -> Sequence:
    # Maybe it can be turned into an object id
    obj_id = db.to_obj_id(arg)
    if obj_id is not None:
        return (obj_id,)

    raise TypeError("Unknown type '{}'".format(arg))


@parse_arg.register(fs.ObjectNode)
def _(arg: fs.ObjectNode):
    return [copy.copy(arg)]


@parse_arg.register(fs.DirectoryNode)
def _(arg: fs.DirectoryNode):
    return [copy.copy(arg)]


@parse_arg.register(fs.ResultsNode)
def _(arg: fs.ResultsNode):
    return parse_args(*arg.children)


@parse_arg.register(os.PathLike)
def _(arg: os.PathLike):
    return [arg]


@parse_arg.register(str)
def _(arg: str):
    obj_id = db.to_obj_id(arg)
    if obj_id is not None:
        return (obj_id,)

    if isinstance(arg, str):
        # Assume it's a path
        return (pathlib.Path(arg),)

    raise TypeError("Unknown type '{}'".format(arg))


def parse_args(*args) -> Sequence:
    parsed = []
    for arg in args:
        parsed.extend(parse_arg(arg))

    return parsed



import datetime
import inspect
import typing
from typing import Mapping


def pretty_type_string(obj_type: typing.Type) -> str:
    """Given an type will return a simple type string"""
    type_str = str(obj_type)
    if type_str.startswith('<class '):
        type_str = type_str[8:-2]

    parts = type_str.split('.')
    if len(parts) > 2:
        return "|".join([parts[0], parts[-1]])

    return type_str


def obj_dict(obj):
    """Given an object return a dictionary that represents it"""
    if isinstance(obj, Mapping):
        return dict(obj)

    repr_dict = {}
    for name in dir(obj):
        if not name.startswith('_'):
            try:
                value = getattr(obj, name)
                if not inspect.isroutine(value):
                    repr_dict[name] = value
            except Exception as exc:  # pylint: disable=broad-except
                repr_dict[name] = '{}: {}'.format(type(exc).__name__, exc)

    return repr_dict


def pretty_datetime(value) -> str:
    if isinstance(value, type):
        return pretty_type_string(value)
    if isinstance(value, datetime.datetime):
        if value.year == datetime.datetime.now().year:
            fmt = "%b %d %H:%M"
        else:
            fmt = "%b %d %Y"

        return value.strftime(fmt)

    return str(value)



from typing import Union, Iterable, Any

import pyos
from pyos import psh


@pyos.psh_lib.command()
def load(*obj_or_ids) -> Union[Iterable[Any], Any]:
    """Load one or more objects"""
    _options, rest = pyos.psh_lib.separate_opts(*obj_or_ids)
    if not rest:
        return None

    to_load = psh.ls(-psh.d, *rest)

    loaded = []
    for node in to_load:
        try:
            loaded.append(node.obj)
        except Exception as exc:  # pylint: disable=broad-except
            loaded.append(exc)

    if len(to_load) == 1:
        return loaded[0]

    return loaded



"""The remove command"""
import argparse
import copy

import cmd2
import tqdm

import pyos
from pyos import psh
from pyos.psh import completion


def _remove_directories(nodes):
    filtered = pyos.fs.ResultsNode()
    for node in nodes:
        if isinstance(node, pyos.fs.DirectoryNode):
            print("rm: cannot remove '{}': Is a directory".format(node.abspath.name))
        elif isinstance(node, pyos.fs.ResultsNode):
            filtered.extend(_remove_directories(node))
        else:
            filtered.append(copy.copy(node))
    return filtered


@pyos.psh_lib.command(pass_options=True)
@pyos.psh_lib.flag(psh.p, help="show progress bar")
@pyos.psh_lib.flag(psh.r, help="remove directories and their contents recursively")
def rm(options, *obj_or_ids):  # pylint: disable=invalid-name
    """Remove objects"""
    to_delete = psh.ls(-psh.d, *obj_or_ids)
    recursive = options.pop(psh.flags.r)
    if not recursive:
        to_delete = _remove_directories(to_delete)

    if to_delete:
        if options.pop(psh.flags.p):
            to_delete = tqdm.tqdm(to_delete, desc="rm")

        for node in to_delete:
            node.delete()


class Rm(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', action='store_true', help="show progress bar")
    parser.add_argument('-r',
                        action='store_true',
                        help="remove directories and their contents recursively")
    parser.add_argument('path', nargs='*', type=str, completer_method=completion.path_complete)

    @cmd2.with_argparser(parser)
    def do_rm(self, app: cmd2.Cmd, args):  # pylint: disable=no-self-use
        command = rm
        if args.r:
            command = command - psh.r
        if args.p:
            command = command - psh.p

        app.poutput(command(*args.path))



import functools
from typing import List

import pytest

from pyos import glob
from pyos import os
from pyos import psh
from . import support


def norm(tempdir, *parts) -> str:
    return os.path.normpath(os.path.join(tempdir, *parts))


def joins(tempdir, *tuples) -> List[str]:
    return [os.path.join(tempdir, *parts) for parts in tuples]


def mktemp(tempdir, *parts):
    filename = norm(tempdir, *parts)
    support.create_empty_file(filename)


change_cwd = support.change_cwd


@pytest.fixture
def test_dir():
    tempdir = "globtest_dir"
    mktemp(tempdir, 'a', 'D')
    mktemp(tempdir, 'aab', 'F')
    mktemp(tempdir, '.aa', 'G')
    mktemp(tempdir, '.bb', 'H')
    mktemp(tempdir, 'aaa', 'zzzF')
    mktemp(tempdir, 'ZZZ')
    mktemp(tempdir, 'EF')
    mktemp(tempdir, 'a', 'bcd', 'EF')
    mktemp(tempdir, 'a', 'bcd', 'efg', 'ha')

    yield tempdir
    # TODO: Remove directory
    psh.rm - psh.r(tempdir)


def check_glob(tempdir, *parts, **kwargs):
    if len(parts) == 1:
        pattern = parts[0]
    else:
        pattern = os.path.join(*parts)
    p = os.path.join(tempdir, pattern)
    res = glob.glob(p, **kwargs)
    assert len(list(glob.iglob(p, **kwargs))) == len(res)
    return res


def assertSequencesEqual_noorder(l1, l2):
    l1 = list(l1)
    l2 = list(l2)
    assert set(l1) == set(l2)
    assert sorted(l1) == sorted(l2)


def test_glob_literal(test_dir):
    eq = assertSequencesEqual_noorder
    do_check = functools.partial(check_glob, test_dir)
    do_norm = functools.partial(norm, test_dir)

    # orig: eq(do_check(test_dir, 'a'), [norm(test_dir, 'a')])
    eq(do_check('a/'), [do_norm('a/')])

    eq(do_check('a', 'D'), [do_norm('a', 'D')])

    # orig: eq(do_check(test_dir, 'aab'), [norm(test_dir, 'aab')])
    eq(do_check('aab/'), [do_norm('aab/')])

    eq(do_check('zymurgy'), [])

    res = check_glob(test_dir, '*')
    assert {type(r) for r in res} == {str}
    res = glob.glob(os.path.join(test_dir, '*'))
    assert {type(r) for r in res} == {str}


def test_glob_one_directory(test_dir):
    eq = assertSequencesEqual_noorder
    do_norm = functools.partial(norm, test_dir)
    do_check = functools.partial(check_glob, test_dir)

    # originals had aaa and aab etc (without '/') as posix doesn't enforce trailing '/' for directories
    eq(do_check('a*'), map(do_norm, ['a/', 'aab/', 'aaa/']))
    eq(do_check('*a'), map(do_norm, ['a/', 'aaa/']))
    eq(do_check('.*'), map(do_norm, ['.aa/', '.bb/']))
    eq(do_check('?aa'), map(do_norm, ['aaa/']))
    eq(do_check('aa?'), map(do_norm, ['aaa/', 'aab/']))
    eq(do_check('aa[ab]'), map(do_norm, ['aaa/', 'aab/']))
    eq(do_check('*q'), [])


def test_glob_nested_directory(test_dir):
    eq = assertSequencesEqual_noorder
    do_norm = functools.partial(norm, test_dir)
    do_check = functools.partial(check_glob, test_dir)

    eq(do_check('a', 'bcd', 'E*'), [do_norm('a/', 'bcd/', 'EF')])
    eq(do_check('a', 'bcd', '*g'), [do_norm('a/', 'bcd/', 'efg/')])


def test_glob_directory_names(test_dir):
    eq = assertSequencesEqual_noorder
    do_norm = functools.partial(norm, test_dir)
    do_check = functools.partial(check_glob, test_dir)

    eq(do_check('*', 'D'), [do_norm('a', 'D')])
    eq(do_check('*', '*a'), [])
    eq(do_check('a', '*', '*', '*a'), [do_norm('a', 'bcd', 'efg', 'ha')])
    eq(do_check('?a?', '*F'), [do_norm('aaa', 'zzzF'), do_norm('aab', 'F')])


def test_glob_directory_with_trailing_slash(test_dir):
    do_norm = functools.partial(norm, test_dir)

    # Patterns ending with a slash shouldn't match non-dirs
    res = glob.glob(do_norm('Z*Z') + os.sep)
    assert res == []
    res = glob.glob(do_norm('ZZZ') + os.sep)
    assert res == []
    # When there is a wildcard pattern which ends with os.sep, glob()
    # doesn't blow up.
    res = glob.glob(do_norm('aa*') + os.sep)
    assert len(res) == 2
    # either of these results is reasonable
    assert set(res) in [
        {do_norm('aaa'), do_norm('aab')},
        {do_norm('aaa') + os.sep, do_norm('aab') + os.sep},
    ]


def check_escape(arg, expected):
    assert glob.escape(arg) == expected


def test_escape():
    check = check_escape
    check('abc', 'abc')
    check('[', '[[]')
    check('?', '[?]')
    check('*', '[*]')
    check('[[_/*?*/_]]', '[[][[]_/[*][?][*]/_]]')
    check('/[[_/*?*/_]]/', '/[[][[]_/[*][?][*]/_]]/')


def rglob(tempdir, *parts, **kwargs):
    return check_glob(tempdir, *parts, recursive=True, **kwargs)


def test_recursive_glob(test_dir):
    eq = assertSequencesEqual_noorder
    full = [
        ('EF',),
        ('ZZZ',),
        ('a/',),
        ('a', 'D'),

        # orig: ('a', 'bcd'),
        ('a', 'bcd/'),
        ('a', 'bcd', 'EF'),

        # orig: ('a', 'bcd', 'efg'),
        ('a', 'bcd', 'efg/'),
        ('a', 'bcd', 'efg', 'ha'),
        ('aaa/',),
        ('aaa', 'zzzF'),
        ('aab/',),
        ('aab', 'F'),
    ]
    eq(rglob(test_dir, '**'), joins(test_dir, ('',), *full))
    eq(rglob(test_dir, os.curdir, '**'),
       joins(test_dir, (os.curdir, ''), *((os.curdir,) + i for i in full)))
    dirs = [('a', ''), ('a', 'bcd', ''), ('a', 'bcd', 'efg', ''), ('aaa', ''), ('aab', '')]

    eq(rglob(test_dir, '**', ''), joins(test_dir, ('',), *dirs))

    eq(
        rglob(test_dir, 'a', '**'),
        joins(test_dir, ('a', ''), ('a', 'D'), ('a', 'bcd/'), ('a', 'bcd', 'EF'),
              ('a', 'bcd', 'efg/'), ('a', 'bcd', 'efg', 'ha')))
    eq(rglob(test_dir, 'a**/'), joins(test_dir, ('a/',), ('aaa/',), ('aab/',)))
    expect = [('a', 'bcd', 'EF'), ('EF',)]

    eq(rglob(test_dir, '**', 'EF'), joins(test_dir, *expect))
    expect = [('a', 'bcd', 'EF'), ('aaa', 'zzzF'), ('aab', 'F'), ('EF',)]

    eq(rglob(test_dir, '**', '*F'), joins(test_dir, *expect))
    eq(rglob(test_dir, '**', '*F', ''), [])
    eq(rglob(test_dir, '**', 'bcd', '*'), joins(test_dir, ('a', 'bcd', 'EF'), ('a', 'bcd', 'efg/')))
    eq(rglob(test_dir, 'a', '**', 'bcd'), joins(test_dir, ('a', 'bcd/')))

    with change_cwd(test_dir):
        join = os.path.join
        eq(glob.glob('**', recursive=True), [join(*i) for i in full])
        eq(glob.glob(join('**', ''), recursive=True), [join(*i) for i in dirs])
        eq(glob.glob(join('**', '*'), recursive=True), [join(*i) for i in full])
        eq(glob.glob(join(os.curdir, '**'), recursive=True),
           [join(os.curdir, '')] + [join(os.curdir, *i) for i in full])
        eq(glob.glob(join(os.curdir, '**', ''), recursive=True),
           [join(os.curdir, '')] + [join(os.curdir, *i) for i in dirs])
        eq(glob.glob(join(os.curdir, '**', '*'), recursive=True),
           [join(os.curdir, *i) for i in full])
        eq(glob.glob(join('**', 'zz*F'), recursive=True), [join('aaa', 'zzzF')])
        eq(glob.glob('**zz*F', recursive=True), [])
        expect = [join('a', 'bcd', 'EF'), 'EF']
        eq(glob.glob(join('**', 'EF'), recursive=True), expect)



"""Change directory command"""
import argparse

import cmd2

import pyos
from pyos.psh import completion


@pyos.psh_lib.command()
def cd(path: pyos.os.PathSpec):  # pylint: disable=invalid-name
    """Change the current working directory"""
    path = pyos.pathlib.PurePath(path)
    if path.is_file_path():
        # Assume they just left out the slash
        path = path.to_dir()

    pyos.os.chdir(path)


class Cd(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('path', nargs=1, type=str, completer_method=completion.dir_completer)

    @cmd2.with_argparser(parser)
    def do_cd(self, args):  # pylint: disable=no-self-use
        cd(args.path[0])



"""This module is for entry points that extend PyOS' functionality"""

import pyos


def get_types():
    """Provide the list of types and helpers we define for mincepy"""
    return (pyos.pathlib.Path,)



import deprecation

from . import pathlib

from .version import __version__

__all__ = ('working_path',)


@deprecation.deprecated(deprecated_in="0.5.0",
                        removed_in="0.6.0",
                        current_version=__version__,
                        details="Use pyos.pathlib.working_path")
def working_path(path: pathlib.Path):
    return pathlib.working_path(path)



"""The tree command"""
import argparse

import cmd2

import pyos
from pyos import psh
from pyos.psh import completion


@pyos.psh_lib.command(pass_options=True)
@pyos.psh_lib.option(psh.flags.L, help='max display depth of the directory tree')
def tree(options, *paths):
    """Get a tree representation of the given paths"""
    to_tree = psh.ls(-psh.d, *paths)
    level = options.pop(psh.L, -1)
    # Fully expand all directories
    for dir_node in to_tree.directories:
        dir_node.expand(level)
    to_tree.show(mode=pyos.fs.TREE_VIEW)
    return to_tree


class Tree(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('-L', type=int, help="max display depth of the directory tree")
    parser.add_argument('path', nargs='*', type=str, completer_method=completion.path_complete)

    @cmd2.with_argparser(parser)
    def do_tree(self, app: cmd2.Cmd, args):  # pylint: disable=no-self-use
        command = tree
        if args.L is not None:
            command = command - psh.L(args.L)

        app.poutput(command(*args.path))



import uuid

import mincepy


class Contact(mincepy.SimpleSavable):
    ATTRS = ('name', 'age', 'tel', 'address')
    TYPE_ID = uuid.UUID('e03a4ca8-ef8f-430a-891c-d58da6f7ad3d')

    def __init__(self, name):
        super().__init__()
        self.name = name
        self.ags = None
        self.tel = None
        self.address = None
        self.friends = mincepy.RefList()



"""Database related classes, functions and constants"""

from .lib import *
from .utils import *
from . import lib
from . import queries

ADDITIONAL = ('queries',)

__all__ = lib.__all__ + utils.__all__ + ADDITIONAL



"""Module containig flags"""

from pyos import psh_lib

__all__ = 'f', 'l', 'L', 'n', 'p', 'd', 'u', 'r', 's'

# pylint: disable=invalid-name

# Define standard flags
d = psh_lib.Option('d')
f = psh_lib.Option('f')
l = psh_lib.Option('l')
L = psh_lib.Option('L')
n = psh_lib.Option('n')
p = psh_lib.Option('p')
r = psh_lib.Option('r')
s = psh_lib.Option('s')
u = psh_lib.Option('u')



import re
from typing import List, Iterator

from pyos import config
from pyos import db
from pyos import os

EXCLUDE_HIDDEN = '(?!\\.)'  # Exclude names starting with dot
MAGIC_CHECK = re.compile('([*?[])')
GLOBSTAR_CHECK = re.compile(r'((?:^|/)\*+(?:$|/))')
PATH_SEGMENT = r'([^/]*)'


def glob(pathname: str, *, recursive=False) -> List[str]:
    return list(iglob(pathname, recursive=recursive))


def iglob(pathname: str, *, recursive=False) -> Iterator[str]:
    hist = db.get_historian()

    if not has_magic(pathname):
        if os.path.exists(pathname):
            yield pathname
        return

    dirname, basename = os.path.split(pathname)

    # Make sure we have an absolute directory (this will also get the cwd if needed)
    dirname = os.path.abspath(dirname)

    if not pathname.endswith(os.sep):
        startdir = _get_startdir(pathname)

        match = GLOBSTAR_CHECK.search(pathname)
        if recursive and match and match.end(0) == len(pathname):
            # Ends in globstar so looking for files AND directories
            found_dirs = set()

            regex = get_dir_regex(os.path.abspath(pathname), recursive=True, complete_match=False)
            regobj = re.compile(regex + r'$')
            query = q_regex(config.DIR_KEY, regex)
            for oid, meta in hist.meta.find(query):
                path = db.get_abspath(oid, meta)

                dirname = os.path.dirname(path)
                yield from get_new_matching_dirs(dirname,
                                                 regobj,
                                                 anchor=startdir,
                                                 existing=found_dirs)

                # Yield the filename
                yield os.path.join(startdir, os.path.relpath(path, startdir))

        else:
            query = {
                **q_regex(config.NAME_KEY, filename_pattern(basename)),
                **q_regex(config.DIR_KEY,
                          get_dir_regex(dirname, recursive=recursive, complete_match=True))
            }

            for oid, meta in hist.meta.find(query):
                path = db.get_abspath(oid, meta)
                yield os.path.join(startdir, os.path.relpath(path, startdir))

            # Check if the pathname ends with magic
            # match = MAGIC_CHECK.search(pathname)
            # if match and match.end(0) == len(pathname):
            # Look for matching directories
            yield from _glob_dirs(pathname, recursive=recursive)
    else:
        # Looking for directories _only_
        yield from _glob_dirs(pathname, recursive=recursive)


def _glob_dirs(dirname: str, *, recursive=False) -> Iterator[str]:
    startdir = _get_startdir(dirname)

    pattern = '^' + get_dir_regex(dirname, recursive=recursive, complete_match=False)

    # Look for objects that are in the given directory or anywhere below
    query = q_regex(config.DIR_KEY, pattern)

    # For each match we have to check that it meets the given criterion as it may be a subdirectory
    regobj = re.compile(pattern + r'$')
    hist = db.get_historian()
    found = set()
    for dname in hist.meta.distinct(config.DIR_KEY, query):
        new = get_new_matching_dirs(dname, regobj, anchor=startdir, existing=found)
        yield from new


def _get_startdir(pathname: str) -> str:
    """Get the starting directory for a pathname search.  This will be the first directory that
    doesn't contain any magic characters"""
    match = MAGIC_CHECK.search(pathname)
    if not match:
        return pathname
    try:
        idx = pathname.rindex(os.sep, 0, match.start(0))
    except ValueError:
        return ''
    else:
        return pathname[:idx + 1]


def q_regex(field: str, pattern: str) -> dict:
    if has_magic(pattern):
        return {field: {'$regex': pattern}}

    return {field: pattern}


def get_dir_regex(pattern: str, *, recursive=False, complete_match=True):
    if not pattern.endswith(os.sep):
        pattern += os.sep
    pattern = os.path.abspath(pattern)

    if not has_magic(pattern):
        return pattern

    regex = get_regex(pattern, recursive=recursive)
    if complete_match:
        return '^' + regex + '$'

    return regex


def filename_pattern(filename: str, complete_match=True) -> str:
    if not filename or not has_magic(filename):
        return filename

    regex = get_regex(filename, recursive=False)
    if complete_match:
        regex = '^' + regex + '$'

    return regex


def get_new_matching_dirs(path: str, regobj, anchor='/', existing: set = None) -> set:
    new = set()
    existing = existing if existing is not None else set()

    relative = os.path.relpath(path, os.path.abspath(anchor))
    if relative == os.curdir:
        partial = anchor
        abspath = os.path.abspath(partial)
        if partial and abspath not in existing and regobj.match(abspath):
            new.add(partial)
            existing.add(abspath)
        return new

    split = relative
    while split and split != os.sep:
        partial = os.path.join(anchor, split)
        abspath = os.path.abspath(partial)
        if abspath not in existing and regobj.match(abspath):
            new.add(partial)
            existing.add(abspath)
        split = os.path.split(split)[0]

    return new


def has_magic(string: str) -> bool:
    return MAGIC_CHECK.search(string) is not None


def escape(pathname: str) -> str:
    """Escape all special characters."""
    return MAGIC_CHECK.sub(r'[\1]', pathname)


def get_regex(pattern: str, recursive=False):
    if not pattern or not has_magic(pattern):
        return pattern

    globstar = recursive

    re_str = ''
    idx = 0
    while idx < len(pattern):
        char = pattern[idx]

        if char in '*?' and _prv(pattern, idx) in [os.sep, None]:
            # We are following from a separator so exclude hidden paths
            re_str += EXCLUDE_HIDDEN

        if char in '/$^+.()=!|':
            re_str += '\\' + char

        elif char == "?":
            re_str += '[^/]'

        elif char == '*':

            # Move over all consecutive "*"'s.
            # Also store the previous and next characters
            pre_char = _prv(pattern, idx)
            star_count = 1

            while _nxt(pattern, idx) == '*':
                star_count += 1
                idx += 1

            next_char = _nxt(pattern, idx)

            if not globstar:
                # globstar is disabled, so treat any number of "*" as one
                re_str += PATH_SEGMENT
            else:
                # globstar is enabled, so determine if this is a globstar segment

                # multiple "*"'s
                # from the start of the segment
                # to the end of the segment

                is_globstar = star_count > 1 \
                              and pre_char in [os.sep, None] \
                              and next_char in [os.sep, None]

                if is_globstar:
                    # it's a globstar, so match zero or more path segments
                    re_str += r'((?:[^/]*(?:\/|$))*)'
                    idx += 1  # move over the "/"
                else:
                    # it's not a globstar, so only match one path segment
                    re_str += PATH_SEGMENT

        else:
            # Default
            re_str += char

        idx += 1

    return re_str


def _nxt(string, idx):
    if idx + 1 >= len(string):
        return None

    return string[idx + 1]


def _prv(string, idx):
    if idx == 0:
        return None
    return string[idx - 1]



"""List command"""
import argparse
import logging

import cmd2

import pyos
from pyos import psh
from pyos import psh_lib
from pyos.psh import completion

logger = logging.getLogger(__name__)


@psh_lib.command(pass_options=True)
@psh_lib.flag(psh.l, help="use a long listing format")
@psh_lib.flag(psh.d, help="list directories themselves, not their contents")
@psh_lib.flag(psh.p, help="print the str() value of each object")
def ls(options, *args) -> pyos.fs.ResultsNode:  # pylint: disable=invalid-name, too-many-branches
    """List the contents of a directory

    :type: restrict listing to a particular type
    """
    rest = args
    parsed = pyos.psh_lib.parse_args(*rest)

    results = pyos.fs.ResultsNode()
    if rest:
        for entry in parsed:
            if isinstance(entry, Exception):
                raise entry

            try:
                results.append(pyos.fs.to_node(entry))
            except ValueError as exc:
                logger.info(str(exc))
    else:
        results.append(pyos.fs.to_node(pyos.pathlib.Path()))

    if not options.pop(psh.d):
        for entry in results:
            if isinstance(entry, pyos.fs.DirectoryNode):
                entry.expand(populate_objects=psh.l in options)

        if len(results) == 1 and isinstance(results[0], pyos.fs.DirectoryNode):
            sole_dir = results[0]
            new_results = pyos.fs.ResultsNode(sole_dir.name)
            for result in tuple(sole_dir.children):
                result.parent = new_results

            results = new_results

    if options.pop(psh.l):
        properties = ['loaded', 'type', 'version', 'mtime', 'name']
        if options.pop(psh.p):
            properties.append('str')
        results.show(*properties, mode=pyos.fs.TABLE_VIEW)
    else:
        results.show(mode=pyos.fs.LIST_VIEW)

    return results


class Ls(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', action='store_true', help="use a long listing format")
    parser.add_argument('-d',
                        action='store_true',
                        help="list directories themselves, not their contents")
    parser.add_argument('-p', action='store_true', help="print the str() value of each object")
    parser.add_argument('path', nargs='*', type=str, completer_method=completion.path_complete)

    @cmd2.with_argparser(parser)
    def do_ls(self, args):  # pylint: disable=no-self-use
        command = ls
        if args.l:
            command = command - psh.l
        if args.d:
            command = command - psh.d
        if args.p:
            command = command - psh.p

        print(command(*args.path))



from . import version
"""

                  ____  ____
                 / __ \/ __ /
    ____  __  __/ / / / /_
   / __ \/ / / / / / /\__ \
  / /_/ / /_/ / /_/ /___/ /
 / .___/\__, /\____//____/
/_/    /____/                 v{}


A new way to interact with your python objects.  PyOS allows you to treat python objects in a 
similar way to the way you interact with files on your filesystem at the moment except in an ipython
console or script.  Objects are stored in a database and presented to the user as existing in a
virtual file system.  Many of the familiar *nix commands are available (ls, mv, rm, tree, cat, find)
except they take on a new, more powerful form because you're in a fully fledged python environment
and not restricted to just two types (file and directories) like on a traditional filesystem but
indeed any python type that can be stored by PyOS's backend.  
""".format(version.__version__)

# Order is important here, first we list all the 'base' modules, then the rest
from . import exceptions
from . import os
from . import db
from . import pathlib
from . import fs

from . import config
from . import fmt
from . import lib
from . import psh
from . import psh_lib

from .version import *
from .lib import *
from .exceptions import *
from .deprecated import working_path
from .pathlib import PurePath, Path

_MODULES = 'os', 'config', 'db', 'fmt', 'fs', 'pathlib', 'psh_lib', 'psh'
_DEPRECATED = ('working_path',)
_ADDITIONAL = ('PurePath', 'Path')

__all__ = version.__all__ + lib.__all__ + exceptions.__all__ + _MODULES + _DEPRECATED + _ADDITIONAL



import argparse
import functools
import logging
import os
import subprocess
import sys
from typing import List, Optional

import click
import mincepy
import cmd2.constants
import cmd2.plugin
import cmd2.utils

import pyos
from pyos import db
from pyos import os as pos
from pyos import version
from . import utils

_LOGGER = logging.getLogger(__name__)

__all__ = 'PyosShell', 'mod'


def mod() -> str:
    """Get the message of the day string"""
    message = [
        version.BANNER,
        "Powered by mincePy (v{})".format(mincepy.__version__),  # pylint: disable=no-member
        'https://pyos.readthedocs.io/en/latest/',
        ''
    ]

    return "\n".join(message)


class PyosShell(cmd2.Cmd):
    """The pyOS shell"""

    def __init__(self,
                 startup_script='',
                 startup_commands: Optional[List[str]] = None,
                 skip_intro=False):
        hist_path = os.path.join(click.get_app_dir('pyos'), 'psh_history')
        super().__init__(startup_script=startup_script,
                         allow_cli_args=False,
                         use_ipython=True,
                         persistent_history_file=hist_path)

        if startup_commands:
            self._startup_commands.extend(startup_commands)

        for cmd_set in utils.plugins_get_commands():
            self.unregister_command_set(cmd_set)

        self.default_to_shell = True

        try:
            pyos.lib.init()
        except (mincepy.ConnectionError, ValueError):
            pass

        if not skip_intro:
            self.intro = mod()
        self._update_prompt()
        self.register_cmdfinalization_hook(self.command_finalise)

    @classmethod
    def execute(cls, *cmd, **kwarg):
        """Run the passed commands and return"""
        init_args = dict(skip_intro=True)
        init_args.update(kwarg)
        cmd = list(cmd) + ['quit']
        init_args['startup_commands'] = cmd
        app = PyosShell(**init_args)
        app.cmdloop()

    def command_finalise(
            self, data: cmd2.plugin.CommandFinalizationData) -> cmd2.plugin.CommandFinalizationData:
        self._update_prompt()
        return data

    def _update_prompt(self):
        try:
            historian = db.get_historian()
        except RuntimeError:
            # Happens when there is a global historian but pyos.db.init() hasn't been called
            historian = None

        if historian is None:
            self.prompt = '[not connected]$ '
        else:
            self.prompt = "{}$ ".format(pos.getcwd())

    def _redirect_output(self, statement: cmd2.Statement) -> cmd2.utils.RedirectionSavedState:
        if statement.pipe_to:
            # Initialize the redirection saved state
            saved = self._create_redirection_save()

            _LOGGER.debug("Attempting piped command:\n"
                          "%s\n"
                          "stdin=%s, stdout=%s", statement.pipe_to, saved.saved_sys_stdin,
                          saved.saved_self_stdout)

            funcs = []
            for part in statement.pipe_to.split(cmd2.constants.REDIRECTION_PIPE):
                # Parse the command
                # Whitespace (especially at beginning) confuses parsing to statement
                part = part.strip()
                part_statement = self._input_line_to_statement(part)
                funcs.append(functools.partial(self.onecmd, part_statement))

            thread_proc = utils.Piper(funcs, out_stream=self.stdout)
            self.stdout = thread_proc.out_redirector
            self.stdin = thread_proc.in_redirector
            saved.redirecting = True

            try:
                self._cur_pipe_proc_reader = thread_proc
                self._redirecting = True
                thread_proc.start()
            except Exception:
                # Shut down the full pipe
                thread_proc.shutdown(wait=True)
                # Restore the default streams
                self._restore_output(statement, saved)
                raise

            return saved

        return super(PyosShell, self)._redirect_output(statement)

    # Preserve quotes since we are passing these strings to the shell
    @cmd2.with_argparser(cmd2.Cmd.shell_parser, preserve_quotes=True)
    def do_shell(self, args: argparse.Namespace) -> None:
        """Execute a command as if at the OS prompt"""
        # Create a list of arguments to shell
        tokens = [args.command] + args.command_args

        # Expand ~ where needed
        cmd2.utils.expand_user_in_tokens(tokens)
        expanded_command = ' '.join(tokens)

        # Prevent KeyboardInterrupts while in the shell process. The shell process will
        # still receive the SIGINT since it is in the same process group as us.
        with self.sigint_protection:
            # For any stream that is a StdSim, we will use a pipe so we can capture its output

            kwargs = dict(
                stdin=sys.stdin if self._redirecting else None,
                # Pass standard input if redirecting
                stdout=subprocess.PIPE
                if isinstance(self.stdout, cmd2.utils.StdSim) else self.stdout,
                stderr=subprocess.PIPE if isinstance(sys.stderr, cmd2.utils.StdSim) else sys.stderr,
                shell=True)

            _LOGGER.debug("Starting shell command: %s. Capturing stdin: %s", expanded_command,
                          self._redirecting)
            proc = subprocess.Popen(expanded_command, **kwargs)

            proc_reader = cmd2.utils.ProcReader(proc, self.stdout, sys.stderr)
            proc_reader.wait()

            # Save the return code of the application for use in a pyscript
            self.last_result = proc.returncode

    def _create_redirection_save(self):
        """Save the current state of the steams and members related to redirection"""
        return utils.RedirectionSavedState(self.stdout, sys.stdout, self.stdin, sys.stdin,
                                           self._cur_pipe_proc_reader, self._redirecting)

    def _restore_output(self, statement: cmd2.Statement,
                        saved_redir_state: utils.RedirectionSavedState):
        """Handles restoring state after output redirection

        :param statement: Statement object which contains the parsed input from the user
        :param saved_redir_state: contains information needed to restore state data
        """
        if saved_redir_state.redirecting:
            # If we redirected output to the clipboard
            if statement.output and not statement.output_to:
                # Fallback to default behaviour
                super(PyosShell, self)._restore_output(statement, saved_redir_state)
            else:
                # Check if we need to wait for the process being piped
                if self._cur_pipe_proc_reader is not None:
                    self._cur_pipe_proc_reader.wait()

                self.stdout = saved_redir_state.saved_self_stdout
                sys.stdout = saved_redir_state.saved_sys_stdout
                self.stdin = saved_redir_state.saved_self_stdin
                sys.stdin = saved_redir_state.saved_sys_stdin
                self._cur_pipe_proc_reader = saved_redir_state.saved_pipe_proc_reader
                self._redirecting = saved_redir_state.saved_redirecting
        else:
            super(PyosShell, self)._restore_output(statement, saved_redir_state)

        if isinstance(saved_redir_state, utils.RedirectionSavedState):
            sys.stdin = saved_redir_state.saved_sys_stdin


if __name__ == '__main__':
    sys.exit(PyosShell().cmdloop())



"""Module for representers which are used to convert objects into strings"""

import functools
import io
import pprint
from typing import Sequence

import mincepy

try:
    import pyprnt
except ImportError:
    pass

from pyos import utils as putils
from pyos import fmt

__all__ = 'get_default', 'set_default', 'dict_representer', 'make_custom_pyprnt', 'prnt'

_DEFAULT_REPRESENTER = None


def get_default():
    global _DEFAULT_REPRESENTER  # pylint: disable=global-statement
    if _DEFAULT_REPRESENTER is None:
        try:
            # We try importing here just in case the user has since installed pyprnt
            import pyprnt as _  # pylint: disable=import-outside-toplevel
            del _
            return prnt
        except ImportError:
            return dict_representer

    return _DEFAULT_REPRESENTER


def set_default(representer):
    global _DEFAULT_REPRESENTER  # pylint: disable=global-statement
    _DEFAULT_REPRESENTER = representer


def dict_representer(obj):
    try:
        if isinstance(obj, mincepy.File):
            return obj.read_text()

        return pprint.pformat(fmt.obj_dict(obj))
    except Exception as exc:  # pylint: disable=broad-except
        return str(exc)


def make_custom_pyprnt(**kwargs):
    return functools.partial(pyprnt.prnt, **kwargs)


def _terminal_width_prnt(obj, **kwargs):
    width = putils.get_terminal_width()
    return pyprnt.prnt(obj, end='', width=width, truncate=True, **kwargs)


def prnt(obj, custom_prnt=None):
    """Pyprnt representer.  Use the great pyprnt library to represent the object either as a
    sequence or a dictionary of the public properties.

    See https://github.com/kevink1103/pyprnt for details of pyprnt"""
    to_prnt = to_simple_repr(obj)
    custom_prnt = custom_prnt or _terminal_width_prnt

    buffer = io.StringIO()
    custom_prnt(to_prnt, output=False, file=buffer)
    return buffer.getvalue()


def to_simple_repr(obj):
    """Convert an object into primitive types i.e. str, list, dict, etc"""
    if isinstance(obj, mincepy.File):
        return obj.read_text()

    if isinstance(obj, Sequence):
        return obj

    if isinstance(obj, Exception):
        return "{}: {}".format(obj.__class__.__name__, obj)

    return fmt.obj_dict(obj)



"""Functions and constants that emulate python's os module"""
from typing import List

import mincepy

from pyos import config
from pyos import db
from pyos import exceptions
from . import path
from . import types

from .path import (curdir, pardir, sep)

# pylint: disable=invalid-name
name = 'pyos'

__all__ = ('getcwd', 'chdir', 'fspath', 'listdir', 'remove', 'sep', 'unlink', 'curdir', 'pardir',
           'path', 'rename', 'scandir', 'DirEntry')


class DirEntry:

    def __init__(self, obj_name: str, obj_path: str, is_file: bool):
        self._name = obj_name
        self._path = obj_path
        self._is_file = is_file

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    def is_dir(self):
        return not self.is_file()

    def is_file(self):
        return self._is_file


def chdir(file_path: types.PathSpec):
    """
    Change the current working directory to path.

    :param file_path: the path to change to
    """
    file_path = path.abspath(file_path)
    if not file_path.endswith(sep):
        file_path += sep
    db.get_historian().meta.sticky[config.DIR_KEY] = file_path


def fspath(file_path: types.PathSpec):
    """Return the pyos representation of the path.

    If str is passed in, it is returned unchanged.
    Otherwise __fspath__() is called and its value is returned as long as it is a str. In all other
    cases, TypeError is raised."""
    if isinstance(file_path, str):
        return file_path

    # Work from the object's type to match method resolution of other magic methods.
    path_type = type(file_path)
    try:
        path_repr = path_type.__fspath__(file_path)
    except AttributeError:
        if hasattr(path_type, '__fspath__'):
            raise

        raise TypeError("expected str or os.PathLike object, not " + path_type.__name__)

    if isinstance(path_repr, str):
        return path_repr

    raise TypeError("expected {}.__fspath__() to return str, not {}".format(
        path_type.__name__,
        type(path_repr).__name__))


def getcwd() -> str:
    """Return a string representing the current working directory."""
    return db.get_historian().meta.sticky.get(config.DIR_KEY, None)


def listdir(lsdir: types.PathSpec = '.') -> List[str]:
    lsdir = path.abspath(lsdir)
    hist = db.get_historian()

    # OBJECTS

    # Query for objects in this directory
    contents = set()
    for obj_id, meta in hist.meta.find(db.path_to_meta_dict(lsdir)):
        contents.add(db.get_obj_name(obj_id, meta))

    # DIRECTORIES

    # Search for objects in directories below this one (depth 1) to any depth (-1) as
    # there may be just folders (no objects) between _abspath and the object in some deeply
    # nested directory
    directories = hist.meta.distinct(config.DIR_KEY, db.queries.subdirs(str(lsdir), 1, -1))

    dirstring_len = len(lsdir)
    for dirstring in directories:
        dirname = dirstring[dirstring_len:]
        contents.add(dirname)

    return list(contents)


# pylint: disable=redefined-builtin
def open(file: types.PathSpec, mode='r', encoding: str = None) -> mincepy.File:
    """
    Open file and return a corresponding file object.

    :param file: the path-like object that gives the pathname to open
    :param mode: the opening mode
    :param encoding: the encoding to use
    :return: a file-like object that can be used to interact with the file
    """
    file_path = path.abspath(file)
    if file_path.endswith(sep):
        raise exceptions.IsADirectoryError(file_path)

    results = db.find_meta(db.path_to_meta_dict(file_path))
    try:
        obj_id = next(results)['obj_id']
        fileobj = db.load(obj_id)
        if not isinstance(file, mincepy.File):
            raise ValueError("open: {}: is not a file".format(fileobj))
    except StopIteration:
        # Create a new one
        fileobj = db.get_historian().create_file(file_path, encoding=encoding)
        db.save_one(fileobj, file_path)

    return fileobj.open(mode=mode)


def remove(file_path: types.PathSpec):
    """Remove (delete) the file path. If path is a directory, an IsADirectoryError is raised. Use
    rmdir() to remove directories."""
    file_path = path.abspath(file_path)
    if file_path.endswith(sep):
        raise exceptions.IsADirectoryError(file_path)

    obj_id = db.get_obj_id(file_path)
    db.get_historian().delete(obj_id)


def rename(src: types.PathSpec, dest: types.PathSpec):
    """Rename the file or directory src to dest.
    If src is a file and dest is a directory or vice-versa, an IsADirectoryError or a
    NotADirectoryError will be raised respectively.
    If both are directories and dest is not empty, an OSError is raised.
    If both are files, dst it will be replaced silently.
    """
    src = fspath(src)
    dest = fspath(dest)
    if src.endswith(sep):
        # Renaming directory
        if not dest.endswith(sep):
            raise exceptions.NotADirectoryError(dest)

        if path.exists(dest):
            # In pyOS a directory must have something in it to exist
            raise OSError("The destination directory '{}' is not empty.".format(dest))

        historian = db.get_historian()
        with historian.transaction():
            # Find everything within this folder at any depth
            query = db.queries.subdirs(src, start_depth=0, end_depth=-1)
            len_src = len(src)
            for obj_id, meta in historian.meta.find(query):
                current = meta[config.DIR_KEY]
                # Update the directory key to contain the new path
                meta[config.DIR_KEY] = path.join(dest, current[len_src:])
                historian.meta.set(obj_id, meta)
    else:
        # Renaming file
        if dest.endswith(sep):
            raise exceptions.IsADirectoryError(dest)

        historian = db.get_historian()
        try:
            obj_id, meta = next(historian.meta.find(db.utils.path_to_meta_dict(src)))
        except StopIteration:
            # Check if the source corresponds to an object id
            obj_id = historian.to_obj_id(path.basename(src))
            if obj_id is None:
                raise exceptions.FileNotFoundError(src)
            meta = historian.meta.get(obj_id)

        # Update the metadata
        meta.update(db.utils.path_to_meta_dict(dest))
        historian.meta.set(obj_id, meta=meta)


def scandir(scan_path='.') -> List[DirEntry]:
    abspath = path.abspath(scan_path)
    hist = db.get_historian()

    # OBJECTS

    # Query for objects in this directory
    contents = {}
    for obj_id, meta in hist.meta.find(db.path_to_meta_dict(abspath)):
        obj_name = db.get_obj_name(obj_id, meta)
        contents[obj_name] = DirEntry(obj_name=obj_name,
                                      obj_path=path.join(scan_path, obj_name),
                                      is_file=True)

    # DIRECTORIES

    # Search for objects in directories below this one (depth 1) to any depth (-1) as
    # there may be just folders (no objects) between _abspath and the object in some deeply
    # nested directory
    directories = hist.meta.distinct(config.DIR_KEY, db.queries.subdirs(str(abspath), 1, -1))

    dirstring_len = len(abspath)
    for dirstring in directories:
        obj_name = dirstring[dirstring_len:]
        if obj_name not in contents:
            contents[obj_name] = DirEntry(obj_name=obj_name,
                                          obj_path=path.join(scan_path, obj_name),
                                          is_file=False)

    return list(contents.values())


def unlink(file_path: types.PathSpec):
    """
    Remove (delete) the file path. This function is semantically identical to remove(); the unlink
    name is its traditional Unix name. Please see the documentation for remove() for further
    information.

    :param file_path: the path to remove
    """
    return remove(file_path)



"""The find command"""
import argparse

import cmd2
from mincepy import qops

import pyos
from pyos.psh import argparse_types
from pyos.psh import completion

__all__ = ('find',)


@pyos.psh_lib.command()
def find(
        *starting_point: pyos.os.PathLike,
        meta: dict = None,
        state: dict = None,
        type=None,  # pylint: disable=redefined-builtin
        mindepth=0,
        maxdepth=-1) -> pyos.fs.ResultsNode:
    """Find objects matching the given criteria, optionally starting from one or more paths


    :param starting_point: one or more paths to start the search from
    :param meta: a dictionary for the metadata to match
    :param state: a dictionary for the record state to match
    :param type: restrict the match to objects of this type
    :param mindepth: the minimum depth to search at relative to the start point(s)
    :param maxdepth: the maximum depth to search at relative to the start point(s)
    """
    _options, spoints = pyos.psh_lib.opts.separate_opts(*starting_point)
    if not spoints:
        spoints = (pyos.pathlib.Path(),)

    return pyos.fs.find(*spoints,
                        meta=meta,
                        state=state,
                        type=type,
                        mindepth=mindepth,
                        maxdepth=maxdepth)


class Find(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('-t',
                        dest='type',
                        type=argparse_types.type_string,
                        help="the type to search for")
    parser.add_argument('-s',
                        dest='paths',
                        action='append',
                        type=str,
                        default=[],
                        completer_method=completion.dir_completer,
                        help="starting point (path) for search")
    parser.add_argument('-m',
                        dest='meta',
                        action='append',
                        type=argparse_types.parse_query,
                        help='a constraint to apply to the metadata')
    parser.add_argument('--maxdepth',
                        dest='maxdepth',
                        type=int,
                        default=-1,
                        help='maximum depth to search at relative to starting point(s) '
                        '(-1 means no maximum)')
    parser.add_argument('--mindepth',
                        dest='mindepth',
                        type=int,
                        default=0,
                        help='minimum depth to start search at relative to starting point(s)')
    parser.add_argument('state',
                        type=argparse_types.parse_query,
                        nargs='*',
                        help='a constraint to apply to the state of the object')

    @cmd2.with_argparser(parser)
    def do_find(self, args):  # pylint: disable=no-self-use
        meta = None
        state = None
        if args.meta:
            meta = qops.and_(*args.meta)
        if args.state:
            state = qops.and_(*args.state)

        res = pyos.fs.find(*args.paths,
                           type=args.type,
                           meta=meta,
                           state=state,
                           mindepth=args.mindepth,
                           maxdepth=args.maxdepth)
        print(res)



from . import db
from . import psh
from . import version
from .psh import *  # pylint: disable=unused-wildcard-import, wildcard-import
from pathlib import PurePath, Path

__all__ = psh.__all__


def _mod() -> str:
    """Get the message of the day string"""
    return "Welcome to\n{}".format(version.BANNER)


print(_mod())
db.init()



"""The move command"""
import argparse

import click
import cmd2

import pyos
from pyos import psh
from pyos.psh import completion


@pyos.psh_lib.command(pass_options=True)
@pyos.psh_lib.flag(psh.f, help="force - do not prompt before overwriting")
def mv(options, *args):  # pylint: disable=invalid-name
    """Take one or more files or directories with the final parameter being interpreted as
     destination

    mv has the following syntax, with the indicated outcomes:

    Files as input:
    mv ('a', 'b/')  - move file 'a' into folder 'b/'
    mv ('a', 'b')   - rename file a to b

    Folders as input:
    mv ('a/', 'b/') - move folder 'a/' into folder 'b/' becoming 'b/a/'
    mv ('a/', 'b')  - rename folder 'a/' to be called 'b' having path 'b/'

    Multiple inputs:
    mv (*args, 'd')     - move all supplied *args to 'd/'
    mv (*args, 'd/')    - move all supplied *args to 'd/'

    For now, files will not be overwritten.
    """
    if len(args) < 2:
        raise ValueError("mv: missing destination")

    args = list(args)
    dest = pyos.Path(args.pop())

    if len(args) > 1:
        # If there is more than one thing to move then we assume that dest is a directory
        dest = dest.to_dir()

    to_move = psh.ls(-psh.d, *args)
    if dest.is_file_path():
        # Renaming
        if dest.exists():
            if -psh.f in options or click.confirm("Overwrite '{}'?".format(dest)):
                dest.unlink()
            else:
                return

        to_move[0].rename(dest.name)
    else:
        dest = dest.resolve()
        to_move.move(dest, overwrite=not options.pop(psh.n))


class Mv(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store_true', help="force - do not prompt before overwriting")
    parser.add_argument('path', nargs='*', type=str, completer_method=completion.path_complete)

    @cmd2.with_argparser(parser)
    def do_mv(self, app: cmd2.Cmd, args):  # pylint: disable=no-self-use
        command = mv
        if args.f:
            command = command - psh.f

        app.poutput(command(*args.path))



"""Library to help writing psh applications and commands"""

from .opts import *
from .representers import *
from .results import *
from .utils import *
from . import opts
from . import results

__all__ = opts.__all__ + representers.__all__ + results.__all__ + utils.__all__



from pyos import os
from pyos import db
from pyos import pathlib

from . import nodes

__all__ = ('find',)


# pylint: disable=redefined-builtin
def find(*starting_point,
         meta: dict = None,
         state: dict = None,
         type=None,
         mindepth=0,
         maxdepth=-1) -> nodes.ResultsNode:
    """
    Find objects matching the given criteria

    :param starting_point: the starting points for the search, if not supplied defaults to '/'
    :param meta: filter criteria for the metadata
    :param state: filter criteria for the object's saved state
    :param type: restrict the search to this type (can be a tuple of types)
    :param mindepth: the minimum depth from the starting point(s) to search in
    :param maxdepth: the maximum depth from the starting point(s) to search in
    :return: results node
    """
    if not starting_point:
        starting_point = (os.getcwd(),)
    meta = (meta or {}).copy()
    state = (state or {}).copy()

    # Converting starting points to abspaths
    spoints = [str(pathlib.PurePath(path).to_dir().resolve()) for path in starting_point]

    # Add the directory search criteria to the meta search
    subdirs_query = (db.queries.subdirs(point, mindepth, maxdepth) for point in spoints)
    meta.update(db.queries.or_(*subdirs_query))

    hist = db.get_historian()

    entries = {}

    if state or type:
        # If we need to match state or type we have to go into the data collection, otherwise
        # we just look up the metadata which is faster

        # Get the records that match both the record and metadata criteria
        for record in hist.find_records(obj_type=type, state=state, meta=meta):
            entries.setdefault(record.obj_id, {})['record'] = record

        # Now get the metadata for those objects
        for obj_id, meta_dict in hist.meta.find(filter={}, obj_id=list(entries.keys())):
            entries.setdefault(obj_id, {})['meta'] = meta_dict
    else:
        for obj_id, meta_dict in hist.meta.find(filter=meta):
            entries.setdefault(obj_id, {})['meta'] = meta_dict

    results = nodes.ResultsNode()
    results.show('relpath')

    for obj_id, entry in entries.items():
        node = nodes.ObjectNode(obj_id,
                                record=entry.get('record', None),
                                meta=entry.get('meta', None))
        results.append(node)

    return results



"""The save command"""

import pyos
from pyos import psh


@pyos.psh_lib.command(pass_options=True)
@pyos.psh_lib.flag(psh.f, help="Force - overwrite files with the same name")
def save(options, *args):
    """Save one or more objects"""
    objs = args

    if len(objs) > 1 and isinstance(objs[-1], (str, pyos.pathlib.PurePath)):
        # Extract the destination
        dest = pyos.pathlib.Path(objs[-1]).resolve()
        objs = objs[:-1]

        if len(objs) > 1 and dest.is_file():
            # Automatically convert to directory if there are many objects as they can't save
            # more than one with the same filename in the same folder!
            dest = dest.to_dir()

        save_args = [(obj, dest) for obj in objs]
    else:
        save_args = [(obj, None) for obj in objs]

    saved = pyos.db.save_many(save_args, overwrite=psh.f in options)

    if len(objs) == 1:
        return saved[0]

    return saved



import abc
from typing import Union

__all__ = 'PathLike', 'PathSpec'


class PathLike(metaclass=abc.ABCMeta):
    """An abstract base class for objects representing a pyos path, e.g. pyos.pathlib.PurePath."""

    @abc.abstractmethod
    def __fspath__(self) -> str:
        """Return the pyos path representation of the object."""


PathSpec = Union[str, PathLike]



"""Print working directory command"""
import argparse
import logging
from typing import Union

import cmd2


class Log(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('logger',
                        help='the logger to set loglevel for, can be empty string for root')
    parser.add_argument('level', help="the log level to set")

    @cmd2.with_argparser(parser)
    def do_log(self, args):  # pylint: disable=no-self-use

        if isinstance(args.level, int):
            level = args.level
        else:
            level = logging.getLevelName(args.level)

        set_logging(args.logger, level)
        print("Log level for '{}' set to {}".format(args.logger, args.level))


def set_logging(logger: str, level: Union[int, str]):
    log = logging.getLogger(logger)
    log.setLevel(level)

    formatter = logging.Formatter('%(levelname)s: %(message)s')

    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(formatter)
    log.addHandler(handler)



"""Module that contains classes used to provide results to the user"""
import collections.abc
from typing import Callable, Iterator

import pyos.results
from pyos.psh_lib import representers

__all__ = 'CachingResults', 'ResultsDict', 'ResultsString'


class CachingResults(collections.abc.Sequence, pyos.results.BaseResults):
    """A helper that takes an iterator and wraps it caching the results as a sequence."""

    def __init__(self, iterator: Iterator, representer: Callable = None):
        """
        Create a caching results sequence.  If no representer is supplied the default will be used.

        :param iterator: the iterable to cache results of
        :param representer: the representer to use, if None the current default will be used.
        """
        super().__init__()
        self._iterator = iterator
        self._representer = representer or representers.get_default()
        self._cache = []

    def __getitem__(self, item):
        self._ensure_cache(item)
        return self._cache[item]

    def __iter__(self):
        return self._iter_generator()

    def __len__(self):
        self._ensure_cache()
        return len(self._cache)

    def __repr__(self):
        return "\n".join([self._representer(item) for item in self])

    def _iter_generator(self, at_end=False):
        idx = 0 if not at_end else len(self._cache)
        while True:
            if idx >= len(self._cache):
                # Ok, try the iterable
                if self._iterator is None:
                    return

                try:
                    self._cache.append(next(self._iterator))
                except StopIteration:
                    self._iterator = None
                    return

            yield self._cache[idx]
            idx += 1

    def _ensure_cache(self, max_index=-1):
        """Fill up the cache up to the max index.  If -1 then fill up entirely"""
        if self._iterator is None or (0 <= max_index < len(self._cache)):
            return

        idx = len(self._cache)
        self_iter = self._iter_generator(at_end=True)
        while True:
            try:
                next(self_iter)
                idx += 1
                if idx == max_index:
                    return
            except StopIteration:
                return


class ResultsDict(collections.abc.Mapping, pyos.results.BaseResults):
    """A custom dictionary representing results from a command"""

    def __init__(self, results: dict, representer=None):
        self._results = results
        self._representer = representer or representers.get_default()

    def __getitem__(self, item):
        return self._results.__getitem__(item)

    def __iter__(self):
        return self._results.__iter__()

    def __len__(self):
        return self._results.__len__()

    def __repr__(self):
        return self._representer(self)


class ResultsString(collections.UserString, pyos.results.BaseResults):
    """A string that overwrites the __repr__ method"""

    def __init__(self, result: str, representer=None):
        super().__init__(result)
        self._representer = representer or representers.get_default()

    def __repr__(self):
        return self._representer(self.data)



import sys

import click

from pyos import psh


@click.command(name='psh')
@click.argument('script', default='')
@click.option('-c', '--cmd', required=False, help="commands to be invoked directly")
def psh_(script, cmd):
    headless_cmds = None
    if cmd:
        headless_cmds = list(part.strip() for part in cmd.split(';'))
    elif script:
        with open(script, 'r') as script_file:
            headless_cmds = list(line.rstrip() for line in script_file.readlines())

    # Need to clear args because otherwise cmd2 picks them up
    if headless_cmds:
        psh.PyosShell.execute(*headless_cmds)
        return

    app = psh.PyosShell()
    sys.exit(app.cmdloop())


@click.group()
def pyos():
    pass


@pyos.command()
def shell():
    """Start the pyos shell"""
    # Need to clear args because otherwise cmd2 picks them up
    app = psh.PyosShell()
    sys.exit(app.cmdloop())



from collections import abc
import getpass
from typing import Sequence, Iterable, Optional, Tuple, Any, Union

import mincepy

from pyos import config
from pyos import exceptions
from pyos import os
from . import utils

__all__ = ('get_historian', 'get_meta', 'update_meta', 'set_meta', 'find_meta', 'save_one',
           'save_many', 'get_abspath', 'load', 'to_obj_id', 'get_obj_id', 'init', 'reset',
           'get_path', 'get_paths', 'rename', 'homedir', 'connect')

_HISTORIAN = None  # type: Optional[mincepy.Historian]


def connect(uri: str = '') -> mincepy.Historian:
    mincepy.connect(uri, use_globally=True)
    return init()


def init():
    # Make sure the indexes are there
    global _HISTORIAN  # pylint: disable=global-statement
    historian = mincepy.get_historian()
    if historian is _HISTORIAN:
        return historian

    historian.meta.create_index([
        (config.NAME_KEY, mincepy.ASCENDING),
        (config.DIR_KEY, mincepy.ASCENDING),
    ],
                                unique=True,
                                where_exist=True)
    historian.meta.create_index(config.NAME_KEY, unique=False, where_exist=True)
    historian.meta.create_index(config.DIR_KEY, unique=False, where_exist=True)

    # Set the current path
    path = '/{}/'.format(getpass.getuser())
    historian.meta.sticky[config.DIR_KEY] = path

    # All done
    _HISTORIAN = historian
    return historian


def get_historian():
    """Get the active historian in pyos"""
    global _HISTORIAN  # pylint: disable=global-statement
    historian = mincepy.get_historian(create=False)
    if historian is not None and historian is not _HISTORIAN:
        raise RuntimeError("PyOS has not been initialised with the current historian.  "
                           "Call pyos.db.init()")
    return _HISTORIAN


def reset():
    global _HISTORIAN  # pylint: disable=global-statement
    if _HISTORIAN is not None:
        _HISTORIAN.meta.sticky.pop(config.DIR_KEY, None)
        _HISTORIAN = None


# region metadata


def get_meta(obj_id: Union[Any, Iterable[Any]]):
    """Get the metadata for a bunch of objects"""
    hist = get_historian()
    return hist.archive.meta_get(obj_id)


def update_meta(*obj_or_identifier: Iterable, meta: dict):
    """Update the metadata for a bunch of objects"""
    hist = get_historian()
    for obj_id in obj_or_identifier:
        hist.meta.update(obj_id, meta)


def set_meta(*obj_or_identifier: Iterable, meta: dict):
    """Set the metadata for a bunch of objects"""
    hist = get_historian()
    obj_ids = tuple(map(to_obj_id, obj_or_identifier))
    metas = dict(find_meta(obj_ids=obj_ids))

    # Preserve the internal keys
    for obj_id in obj_ids:
        new_meta = utils.new_meta(metas.get(obj_id, {}), meta)
        hist.meta.set(obj_id, new_meta)


def find_meta(filter: dict = None, obj_ids=None):  # pylint: disable=redefined-builtin
    filter = filter or {}
    hist = get_historian()
    return hist.meta.find(filter, obj_ids)


# endregion

# region paths


def get_path(obj_or_id) -> Optional[str]:
    """Given an object or object id get the current path"""
    return get_paths(obj_or_id)[0]


def get_paths(*obj_or_id) -> Sequence[Optional[str]]:
    """Given objects or identifier this will return their current paths in the order they were
    passed"""
    hist = get_historian()
    obj_ids = tuple(map(hist.get_obj_id, obj_or_id))
    metas = dict(hist.meta.find({}, obj_id=obj_ids))
    paths = []
    for obj_id in obj_ids:
        meta = metas.pop(obj_id)
        dirname, name = meta.get(config.DIR_KEY, None), meta.get(config.NAME_KEY, None)
        path = []
        if dirname:
            path.append(dirname)
        path.append(name or str(obj_id))

        paths.append(os.path.join(*path))

    return paths


def rename(obj_or_id, dest: os.PathSpec):
    """Rename an object to a the dest.  If dest is is a directory IsADirectoryError is raised."""
    dest = os.fspath(dest)
    if dest.endswith(os.sep):
        raise exceptions.IsADirectoryError(dest)

    hist = get_historian()
    obj_id = hist.to_obj_id(obj_or_id)
    hist.meta.update(obj_id, meta=utils.path_to_meta_dict(dest))


def get_abspath(obj_id, meta: dict) -> str:
    """Given an object id and metadata dictionary this method will return a string representing
    the absolute path of the object"""
    assert obj_id, "Must provide a valid obj id"
    dirname = meta[config.DIR_KEY]
    basename = meta.get(config.NAME_KEY, str(obj_id))
    return os.path.join(dirname, basename)


def homedir() -> str:
    """Return the users home directory"""
    user_info = get_historian().get_user_info()
    user_name = user_info[mincepy.ExtraKeys.USER]
    return "/{}/".format(user_name)


# endregion


def save_one(obj, path: os.PathSpec = None, overwrite=False, historian=None):
    """Save one object to a the given path.  The path can be a filename or a directory or a filename
    in a directory

    :param obj: the object to save
    :param path: the optional path to save it to
    :param overwrite: overwrite if there is already an object at that path
    """
    hist = historian or get_historian()

    if path is None:
        # Check if it's already saved (in which case we don't need to set a path)
        if hist.get_obj_id(obj) is None:
            path = os.getcwd()

    meta = None
    if path is not None:
        path = os.path.abspath(path)
        meta = utils.path_to_meta_dict(path)

    try:
        return hist.save_one(obj, meta=meta)
    except mincepy.DuplicateKeyError:
        if overwrite:
            os.remove(path)
            # Try again
            return hist.save_one(obj, meta=meta)

        raise


def save_many(to_save: Iterable[Union[Any, Tuple[Any, os.PathSpec]]],
              overwrite=False,
              historian=None):
    """
    Save many objects, expects an iterable where each entry is an object to save or a tuple of
    length 2 containing the object and a path of where to save it.

    :param to_save: the iterable able objects to save
    :param overwrite: overwrite objects with the same name
    """
    obj_ids = []
    for entry in to_save:
        path = None
        if isinstance(entry, abc.Sequence):
            if len(entry) > 2:
                raise ValueError("Can only pass sequences of at most length 2")
            obj, path = entry[0], entry[1]
        else:
            # Assume it's just the object
            obj = entry
        obj_ids.append(save_one(obj, path, overwrite, historian=historian))

    return obj_ids


def load(*identifier):
    """Load one or more objects"""
    get_historian().load(*identifier)


def to_obj_id(identifier):
    """Get the database object id from the passed identifier.  If the identifier is already a
    mincepy object id it will be returned unaltered.  Otherwise mincepy will try and turn the type
    into an object id.  It it fails, None is returned"""
    return get_historian().to_obj_id(identifier)


def get_obj_id(path: os.PathSpec):
    """Given a path get the id of the corresponding object.  Returns None if not found."""
    results = find_meta(utils.path_to_meta_dict(path))
    try:
        return next(results)[0]
    except StopIteration:
        return None



"""The meta command"""
import argparse
import logging
import sys

import cmd2

import pyos
from pyos.psh import completion
from pyos import psh

logger = logging.getLogger(__name__)


@pyos.psh_lib.command(pass_options=True)
@pyos.psh_lib.flag(psh.s, "Set the metadata")
@pyos.psh_lib.flag(psh.u, "Update the metadata")
def meta(options, *obj_or_ids, **updates):
    """Get, set or update the metadata on one or more objects"""
    if not obj_or_ids:
        return None

    to_update = psh.ls(-psh.d, *obj_or_ids)
    obj_ids = []
    for node in to_update:
        if isinstance(node, pyos.fs.ObjectNode):
            obj_ids.append(node.obj_id)
        else:
            print("Can't set metadata on '{}'".format(node))

    if options.pop(psh.u, False):
        # In 'update' mode
        if not updates:
            return None
        pyos.db.lib.update_meta(*obj_ids, meta=updates)
    elif options.pop(psh.s, False):
        # In 'setting' mode
        if not updates:
            return None
        pyos.db.lib.set_meta(*obj_ids, meta=updates)
    else:
        # In 'getting' mode
        if updates:
            logging.warning("Keywords supplied to meta without -s/-u flags: %s", updates)

        if len(obj_ids) == 1:
            # Special case for a single parameter
            return pyos.psh_lib.ResultsDict(pyos.db.lib.get_meta(obj_ids[0]))

        return pyos.psh_lib.CachingResults(pyos.db.lib.find_meta(obj_ids=obj_ids))

    return None


class Meta(cmd2.CommandSet):
    parser = argparse.ArgumentParser()

    set_update = parser.add_mutually_exclusive_group()
    set_update.add_argument('-s', action='store_true', help="set the metadata")
    set_update.add_argument('-u', action='store_true', help="update the metadata")
    parser.add_argument('path', nargs='*', type=str, completer_method=completion.path_complete)

    @cmd2.with_argparser(parser)
    def do_meta(self, args):  # pylint: disable=no-self-use
        command = meta
        if args.u:
            command = command - psh.u
        if args.s:
            command = command - psh.s

        if not args.path:
            # Read from standard in
            try:
                args.path = [line.rstrip() for line in sys.stdin.readlines()]
            except Exception:
                logger.exception("Exception trying to readlines")
                raise

        logger.debug("Writing to %s", sys.stdout)
        result = command(*args.path)
        rep = repr(result)
        print(rep, file=sys.stdout)



from pyos.db import queries
from .flags import *
from .cmds import *
from .shell import *
from . import flags
from . import cmds
from . import completion
from . import shell

cwd = completion.PathCompletion('.')  # pylint: disable=invalid-name

_ADDITIONAL = 'queries', 'completion', 'cwd'

__all__ = cmds.__all__ + flags.__all__ + shell.__all__ + _ADDITIONAL



"""Print working directory command"""
import argparse

import cmd2

import pyos


@pyos.psh_lib.command()
def pwd() -> pyos.pathlib.Path:
    """Return the current working directory"""
    return pyos.pathlib.Path().resolve()


class Pwd(cmd2.CommandSet):
    parser = argparse.ArgumentParser()

    @cmd2.with_argparser(parser)
    def do_pwd(self, app, _):  # pylint: disable=no-self-use
        app.poutput(pwd())



from pyos import config
from pyos import os

__all__ = ('path_to_meta_dict', 'get_obj_name')


def new_meta(orig: dict, new: dict) -> dict:
    merged = new.copy()
    if not orig:
        return merged

    for name in config.KEYS:
        if name in orig:
            if name.startswith('_'):
                # Always take internal, i.e. underscored, keys
                merged[name] = orig[name]
            else:
                merged.setdefault(name, orig[name])

    return merged


def path_to_meta_dict(path: os.PathSpec) -> dict:
    """
    :param path: the path to get a dictionary for
    :return: the meta dictionary with the path
    """
    if path is None:
        return {}

    path = os.path.normpath(path)
    meta = {}
    if path.endswith(os.sep):
        meta[config.DIR_KEY] = path
    else:
        dirname, basename = os.path.split(path)
        meta[config.NAME_KEY] = basename
        if dirname:
            meta[config.DIR_KEY] = os.path.abspath(dirname)

    return meta


def get_obj_name(obj_id, meta: dict) -> str:
    """Get the name of an object.  This will be the name that is used to represent this object on
    the virtual filesystem and is stored in the metadata"""
    return meta.get(config.NAME_KEY, str(obj_id))



"""Pyos versions os python's os.path methods"""
import posixpath
from typing import Iterable, Tuple

from pyos import config
from pyos import db
from pyos.os import pos
from . import types

# pylint: disable=invalid-name

sep = '/'  # The path separator pylint: disable=invalid-name
curdir = '.'  # Used to refer to the current directory
pardir = '..'  # Used to refer to the parent directory


def isabs(path: types.PathSpec):
    """Test whether a path is absolute

    :type path: pyos.os.PathSpec
    """
    return pos.fspath(path).startswith(sep)


def abspath(path: types.PathSpec) -> str:
    """Return a normalised absolutised version of the pathname path. This is equivalent to calling
    the function normpath() as follows: normpath(join(os.getcwd(), path)).
    """
    path = pos.fspath(path)
    if isabs(path):
        return path

    return normpath(join(pos.getcwd(), path))


def join(path: types.PathSpec, *paths: Iterable[types.PathSpec]) -> str:
    """Join one or more path components intelligently. The return value is the concatenation of path
    and any members of *paths with exactly one directory separator (os.sep) following each non-empty
    part except the last, meaning that the result will only end in a separator if the last part is
    empty. If a component is an absolute path, all previous components are thrown away and joining
    continues from the absolute path component.
    """
    return posixpath.join(pos.fspath(path), *[pos.fspath(entry) for entry in paths])


def normpath(path: types.PathSpec) -> str:
    """Normalise a pathname by collapsing redundant separators and up-level references so that A//B,
    A/./B and A/foo/../B all become A/B.  The standard pyos distinction between object paths and
    directories is maintained to paths ending with a '/' will be kept as such.

    :type path: pyos.os.PathSpec
    """
    dot = '.'
    path = pos.fspath(path)
    needs_final_slash = path.endswith(sep) or path.endswith(dot)
    path = posixpath.normpath(path)
    if path.startswith('//'):
        path = path[1:]
    # Posix will never return a trailing '/' so check if we need to add it back
    if needs_final_slash and path != '/':
        path += sep

    return path or dot + sep


def basename(path: types.PathSpec) -> str:
    """Return the base name of pathname path. This is the second element of the pair returned by
    passing path to the function split(). Note that the result of this function is different from
    the Unix basename program; where basename for '/foo/bar/' returns 'bar', the basename() function
    returns an empty string ('').
    """
    return posixpath.basename(pos.fspath(path))


def dirname(path: types.PathSpec) -> str:
    """Return the directory name of pathname path. This is the first element of the pair returned by
    passing path to the function split()."""
    name = posixpath.dirname(pos.fspath(path))
    if name and not name.endswith(sep):
        name += sep

    return name


def exists(path: types.PathSpec):
    """Return `True` if the path exists"""
    path = abspath(path)
    if path.endswith(sep):
        query = {config.DIR_KEY: {'$regex': r'^{}.*'.format(path)}}
    else:
        query = db.path_to_meta_dict(path)
    results = db.find_meta(query)
    try:
        # Check if we have hits
        next(results)
        return True
    except StopIteration:
        return False


def expanduser(path: types.PathSpec) -> str:
    """An initial `~` is replaced with the users home directory which is '/[username]'/."""
    homedir = db.homedir()
    path = pos.fspath(path)
    return normpath(path.replace('~', homedir, 1))


def split(path: types.PathSpec) -> Tuple[str, str]:
    """Split the pathname path into a pair, (head, tail) where tail is the last pathname component
    and head is everything leading up to that.  If there is no slash in path, head will be empty. If
    path is empty, both head and tail are empty.
    In all cases, join(head, tail) returns a path to the same location as path (but the strings may
    differ).
    """
    if path == sep:
        return sep, ''

    if path.endswith(sep):
        idx = path[:-1].rfind(sep)
    else:
        idx = path.rfind(sep)

    head, tail = path[:idx + 1], path[idx + 1:]
    if head and head != sep * len(head):
        head = head.rstrip(sep)
        head += sep

    return head, tail


def relpath(path: types.PathSpec, start=curdir) -> str:
    """
    Return a relative filepath to path either from the current directory or from an optional start
    directory. This is a path computation: the filesystem is not accessed to confirm the existence
    or nature of path or start.

    start defaults to pos.curdir.

    :param path: the path to get the relative path for
    :param start: the start directory
    """
    if not path:
        raise ValueError("no path specified")

    path = pos.fspath(path)
    is_dir = path.endswith(sep)

    if start is None:
        start = curdir
    else:
        start = pos.fspath(start)

    try:
        start_list = [x for x in abspath(start).split(sep) if x]
        path_list = [x for x in abspath(path).split(sep) if x]
        # Work out how much of the filepath is shared by start and path.
        i = len(commonprefix([start_list, path_list]))

        rel_list = [pardir] * (len(start_list) - i) + path_list[i:]
        if not rel_list:
            return curdir
        if is_dir:
            rel_list[-1] += sep
        return join(*rel_list)
    except (TypeError, AttributeError, DeprecationWarning):
        _check_arg_types('relpath', path, start)
        raise


# Return the longest prefix of all list elements.
def commonprefix(names):
    """Given a list of pathnames, returns the longest common leading component"""
    if not names:
        return ''
    if not isinstance(names[0], (list, tuple)):
        names = tuple(map(pos.fspath, names))
    str1 = min(names)
    str2 = max(names)
    for i, char in enumerate(str1):
        if char != str2[i]:
            return str1[:i]
    return str1


def isdir(path: types.PathSpec) -> bool:
    """Return True if path is an existing directory."""
    meta = db.path_to_meta_dict(abspath(path))
    try:
        db.get_historian().meta.find(meta)
        return True
    except StopIteration:
        return False


def _check_arg_types(funcname, *args):
    for arg in args:
        if not isinstance(arg, str):
            raise TypeError('{}() argument must be str, bytes, or '
                            'os.PathLike object, not {}'.format(funcname,
                                                                arg.__class__.__name__)) from None



import collections
import concurrent.futures
import contextlib

try:
    from contextlib import nullcontext
except ImportError:
    from contextlib2 import nullcontext
import io
import logging
import os
import sys
import threading
import traceback
from typing import Optional, Union, TextIO, List, Callable, Tuple

import cmd2.utils
import stevedore

_LOGGER = logging.getLogger(__name__)


class ThreadStreamRedirector(io.TextIOBase):
    """Idea for this was found here:
    https://stackoverflow.com/questions/14890997/redirect-stdout-to-a-file-only-for-a-specific-thread
    """

    def __init__(self, *, default: Optional[TextIO] = None, name=''):
        super().__init__()
        self.default = default
        self._streams = {}
        self._name = name

    def register(self, stream: Optional[TextIO]):
        identity = threading.current_thread().ident
        _LOGGER.debug("(%s): Redirecting thread %s to %s", self._name, identity, stream)
        self._streams[identity] = stream

    def unregister(self):
        identity = threading.current_thread().ident
        _LOGGER.debug("(%s): Unregistering thread %s", self._name, identity)
        self._streams.pop(identity)

    @contextlib.contextmanager
    def redirect(self, stream: Optional[TextIO]):
        self.register(stream)
        yield
        self.unregister()

    def __iter__(self):
        return self._get_stream().__iter__()

    def __next__(self):
        return self._get_stream().__next__()

    @property
    def closed(self) -> bool:
        return self._get_stream().closed

    def close(self):
        if getattr(self, '_finalizing', False):
            # We're being asked to close because of finalising but because we don't own any streams
            # ourselves, we just redirect, we are not responsible for closing them in this case.
            return

        stream = self._get_stream()
        logging.debug("Closing stream %s", stream)
        if stream.name == '<stdin>':
            logging.warning("Someone is closing the standard input: \n%s",
                            ''.join(traceback.format_stack()))
        stream.close()

    def write(self, message):
        stream = self._get_stream()
        logging.debug("Writing %s from thread %s to %s", message,
                      threading.current_thread().ident, stream)
        stream.write(message)

    def read(self, size=-1):
        return self._get_stream().read(size=size)

    def readline(self):
        return self._get_stream().readline()

    def readlines(self):
        return self._get_stream().readlines()

    def flush(self):
        self._get_stream().flush()

    def seek(self, offset, whence=io.SEEK_SET):
        return self._get_stream().seek(offset, whence=whence)

    def tell(self):
        return self._get_stream().tell()

    def isatty(self):
        return self._get_stream().isatty()

    def fileno(self):
        return self._get_stream().fileno()

    def _get_stream(self):
        identity = threading.current_thread().ident
        try:
            return self._streams[identity]
        except KeyError:
            return self.default


class Piper:

    def __init__(self, funcs: List[Callable], out_stream: TextIO = sys.stdout):
        self._funcs = funcs
        self._out_stream = out_stream

        self._pipes = {}
        self._in_steam = None
        self._thread_pool = None
        self._orig_stdin = sys.stdin
        self._orig_stdout = sys.stdout

        # Open up the 0th pipe and set up our input stream
        _read, write = self._get_pipe(0)
        self._in_steam = open(write, 'w')

        # Input/Output redirectors
        self._in_redir = ThreadStreamRedirector(name='stdin', default=self._orig_stdin)
        self._out_redir = ThreadStreamRedirector(name='stdout', default=self._orig_stdout)

    @property
    def in_stream(self) -> TextIO:
        return self._in_steam

    @property
    def in_redirector(self) -> TextIO:
        return self._in_redir

    @property
    def out_redirector(self) -> TextIO:
        return self._out_redir

    def start(self, capture_this_stdout=True, capture_all_stdout=False):
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(256)
        try:
            # Start redirecting standard streams
            sys.stdin = self._in_redir
            sys.stdout = self._out_redir

            if capture_this_stdout:
                # On this thread, redirect standard out to our input stream
                self._out_redir.register(self._in_steam)
            if capture_all_stdout:
                self._out_redir.default = self._in_steam

            for idx, func in reversed(list(enumerate(self._funcs))):
                # Create a pipe at the INPUT side of this func
                read, _write = self._get_pipe(idx)

                # OUTPUT: Determine what the output stream for this part should be
                if idx == len(self._funcs) - 1:
                    # At the end, so just go to standard out
                    open_out_stream = nullcontext(self._out_stream)
                else:
                    # Need to output to the next commands input
                    _next_read, next_write = self._pipes[idx + 1]
                    open_out_stream = open(next_write, 'w')

                done_redirecting = threading.Event()

                future = self._thread_pool.submit(self._run_func, func, open(read, 'r'),
                                                  open_out_stream, done_redirecting)

                try:
                    done_redirecting.wait(timeout=0.1)
                except TimeoutError:
                    # Try checking the future first
                    future.result(timeout=0.)

                    # Didn't raise, so we need to raise ourselves
                    raise RuntimeError("Failed to redirect streams for command '{}' in a timely "
                                       "manner".format(func))
                else:
                    del done_redirecting

        except Exception:
            self.shutdown(wait=True)
            raise

    def shutdown(self, wait=True):
        """Shutdown the pipe processor and clean up"""
        if self._thread_pool is None:
            return

        _LOGGER.debug("Shutting down pipe processor")

        # First thing is to close the input stream as this will cause it to flush
        # and downstream commands may still be waiting for input
        try:
            self._in_steam.close()
        except BrokenPipeError:
            _LOGGER.debug("Broken pipe error")

        _LOGGER.debug("Waiting for thread pool to shut down")
        self._thread_pool.shutdown(wait=wait)
        if self._thread_pool is None:
            # Another thread already beat us to it
            return

        _LOGGER.debug("Thread pool shut down")

        self._in_steam = None
        sys.stdin = self._orig_stdin
        sys.stdout = self._orig_stdout
        self._orig_stdin = None
        self._orig_stdout = None
        self._in_redir = None
        self._out_redir = None

        self._funcs = None
        self._pipes = None
        self._thread_pool = None

    # region Cmd2 compatibility
    def send_sigint(self):
        """Send a SIGINT to the process similar to if <Ctrl>+C were pressed"""
        self.shutdown(wait=False)

    def terminate(self):
        """Terminate the process"""
        self.shutdown(wait=False)

    def wait(self):
        """Wait for the process to finish"""
        self.shutdown()

    # endregion

    def _run_func(self, func, open_in_stream, open_out_stream, done_redirecting: threading.Event):
        """
        Run a function that is part of the pipe.

        :param func: the function to invoke.  Will be called without any arguments
        :param open_in_stream: the input stream for the function
        :param open_out_stream: the output stream for the function
        :param done_redirecting: an event that tells the caller that redirection has been set up
        :return: the return value of the func()
        """
        with open_in_stream as in_stream, open_out_stream as out_stream:
            with self._in_redir.redirect(in_stream), self._out_redir.redirect(out_stream):
                done_redirecting.set()
                return func()

    def _get_pipe(self, idx) -> Tuple:
        try:
            pipe = self._pipes[idx]
        except KeyError:
            pipe = os.pipe()
            self._pipes[idx] = pipe
            _LOGGER.debug("Created pipe %s for func %s", pipe, self._funcs[idx])

        return pipe


PipeInfo = collections.namedtuple('PipeInfo', 'pipe_in pipe_out')


class RedirectionSavedState(cmd2.utils.RedirectionSavedState):
    """Extension of cmd2's standard redirection saved state to support additional properties"""

    # pylint: disable=too-few-public-methods

    # pylint: disable=too-many-arguments
    def __init__(self, self_stdout: Union[cmd2.utils.StdSim,
                                          TextIO], sys_stdout: Union[cmd2.utils.StdSim, TextIO],
                 self_stdin: TextIO, sys_stdin: TextIO,
                 pipe_proc_reader: Optional[cmd2.utils.ProcReader], saved_redirecting: bool):
        super().__init__(self_stdout, sys_stdout, pipe_proc_reader, saved_redirecting)
        self.saved_self_stdin = self_stdin
        self.saved_sys_stdin = sys_stdin


PLUGINS_COMMANDS_NS = 'pyos.plugins.shell'


def plugins_get_commands() -> List:
    """Get all mincepy types and type helper instances registered as extensions"""
    mgr = stevedore.extension.ExtensionManager(
        namespace=PLUGINS_COMMANDS_NS,
        invoke_on_load=False,
    )

    commands = []

    def get_command(extension: stevedore.extension.Extension):
        try:
            commands.extend(extension.plugin())
        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception("Failed to get command plugin from %s", extension.name)

    try:
        mgr.map(get_command)
    except stevedore.extension.NoMatches:
        pass

    return commands



"""The locate command"""
import argparse
from typing import Optional, Union, Sequence

import cmd2

import pyos
from pyos import psh


@pyos.psh_lib.command()
def locate(*obj_or_ids) -> Optional[Union[pyos.pathlib.Path, Sequence[pyos.pathlib.Path]]]:
    """Locate the directory of one or more objects"""
    if not obj_or_ids:
        return None

    to_locate = psh.ls(-psh.d, *obj_or_ids)
    # Convert to abspaths
    paths = [node.abspath for node in to_locate]
    results = pyos.psh_lib.CachingResults(paths.__iter__(), representer=str)

    if len(obj_or_ids) == 1 and len(results) == 1:
        return results[0]

    return results


class Locate(cmd2.CommandSet):
    parser = argparse.ArgumentParser()
    parser.add_argument('obj_ids', nargs='*', type=str)

    @cmd2.with_argparser(parser)
    def do_locate(self, app: cmd2.Cmd, args):  # pylint: disable=no-self-use
        app.poutput(locate(*args.obj_ids))



import argparse
import logging
import sys

import cmd2

import pyos
from pyos import db
from pyos import psh
from pyos.psh import completion

_LOGGER = logging.getLogger(__name__)


@pyos.psh_lib.command()
def cat(*obj_or_ids, representer=None):
    """Convert the contents of objects into strings.
    A representer can optionally be passed in which should take the passed object and convert it to
    a string.
    """
    if not obj_or_ids:
        return None

    hist = db.get_historian()
    to_cat = []

    for entry in obj_or_ids:
        if isinstance(entry,
                      (str, pyos.pathlib.Path, pyos.fs.BaseNode, hist.archive.get_id_type())):
            to_cat.extend(psh.ls(-psh.d, entry))
        else:
            to_cat.append(entry)

    representer = representer or pyos.psh_lib.get_default()

    def iterator():
        for node in to_cat:
            try:
                if isinstance(node, pyos.fs.DirectoryNode):
                    yield "cat: {}: Is a directory".format(node.abspath.name)
                elif isinstance(node, pyos.fs.ObjectNode):
                    yield representer(node.obj)
                else:
                    yield representer(node)
            except Exception as exc:  # pylint: disable=broad-except
                yield representer(exc)

    results = pyos.psh_lib.CachingResults(iterator(), representer=str)

    if len(to_cat) == 1:
        return pyos.psh_lib.ResultsString(results[0])

    return results


class Cat(cmd2.CommandSet):
    ls_parser = argparse.ArgumentParser()
    ls_parser.add_argument('path', nargs='*', type=str, completer_method=completion.file_completer)

    @cmd2.with_argparser(ls_parser)
    def do_cat(self, args):  # pylint: disable=no-self-use
        command = cat
        if not args.path:
            # Read from standard in
            _LOGGER.debug("cat: getting input from stdin")
            try:
                args.path = [line.rstrip() for line in sys.stdin.readlines()]
            except Exception:
                _LOGGER.exception("Exception trying to readlines")
                raise
            _LOGGER.debug("cat: got input' %s' from stdin", args.path)

        print(command(*args.path))



from typing import Callable


class BaseResults:

    def __or__(self, other: Callable):
        if not isinstance(other, Callable):
            raise ValueError("'{}' is not callable".format(other))

        return other(self)



import abc
import functools
from typing import Callable
import types

__all__ = 'command', 'Option', 'Options', 'separate_opts', 'option', 'flag'


class Option:
    """An option that can be accepted by a command"""

    def __init__(self, name, call_params=()):
        self._name = name
        self._params_stack = call_params

    def __str__(self):
        parts = ["-", self.name]
        for params in self._params_stack:
            args = params[0]
            kwargs = params[1]

            params = list(str(arg) for arg in args)
            params.extend(tuple("{}={}".format(key, value) for key, value in kwargs.items()))
            parts.extend(('(', ",".join(params), ')'))

        return "".join(parts)

    def __call__(self, *args, **kwargs):
        if len(self._params_stack) == 2:
            raise RuntimeError("Option already has filled parameters")

        params = list(self._params_stack)
        params.append((args, kwargs))
        return Option(self.name, tuple(params))

    def __neg__(self):
        """Allow so a user can optional use a dash when passing this as an argument"""
        return self

    @property
    def called(self) -> bool:
        return bool(self._params_stack)

    @property
    def params_stack(self) -> tuple:
        return self._params_stack

    @property
    def name(self):
        return self._name

    def new(self):
        """Get a new option of this type"""
        return Option(self.name)


class OptionSpec:

    def __init__(self, opt: Option, is_flag=True, help=''):  # pylint: disable=redefined-builtin
        self.option = opt
        self.is_flag = is_flag
        self.help = help


class Options:
    """Container storing options value, using this with separate_opts makes writing tools much
     easier"""

    def __init__(self):
        self._opts = {}  # typing.MutableMapping[str, Option]

    def __str__(self):
        return " ".join((str(val) for val in self._opts.values()))

    def __contains__(self, item):
        if isinstance(item, Option):
            return item.name in self._opts
        if isinstance(item, str):
            return item in self._opts

        return False

    def add(self, name, value):
        self._opts[name] = value

    def pop(self, opt: Option, *default):
        if not isinstance(opt, Option):
            raise TypeError("Unsupported option type '{}'".format(opt))
        if len(default) > 1:
            raise ValueError("Can only pass at most one default")

        name = opt.name
        return self._opts.pop(name, default[0] if default else None)

    def update(self, other):
        for name, opt in other._opts.items():  # pylint: disable=protected-access
            if name not in self._opts:
                self._opts[name] = opt
            else:
                raise ValueError("Option '{}' alreday specified!".format(name))


def separate_opts(*args) -> [Options, list]:
    """Separate out the options from the other arguments"""
    opts = Options()
    rest = []
    for arg in args:
        if isinstance(arg, Option):
            params_stack = arg.params_stack
            value = True
            if params_stack and params_stack[0] and params_stack[0][0]:
                value = params_stack[0][0][0]

            opts.add(arg.name, value)
        elif isinstance(arg, Options):
            opts.update(arg)
        else:
            rest.append(arg)
    return opts, rest


# region Commands


def command(pass_options=False):

    def wrapper(func):
        if isinstance(func, Command):
            func.pass_options = pass_options
            return func

        return Command(func, pass_options)

    return wrapper


class CommandLike(metaclass=abc.ABCMeta):
    """Base class for command like things"""

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        """Call this command"""

    def __or__(self, other: Callable):
        """Pipe the result of this call into a callable

        :param other: the callable to pipe the results to
        """
        result = self.__call__()
        return result.__or__(other)


class Command(CommandLike):
    """Class representing a command"""

    def __init__(self, func: callable, pass_options=False):
        self.func = func
        self.pass_options = pass_options
        self.accepts = {}

        # Use partial just to create a new function instance so we can set the docstring on this
        # instance only
        self.__call__ = types.MethodType(functools.partial(Command.__call__), self)
        self._create_docstring()

    @property
    def name(self) -> str:
        return self.func.__name__

    def add_opt(self, spec: OptionSpec):
        """Add an option that this command accepts"""
        self.accepts[spec.option.name] = spec
        self._create_docstring()

    def __call__(self, *args, **kwargs):  # pylint: disable=method-hidden
        """Execute this command directly"""
        if self.pass_options:
            options, rest = separate_opts(*args)
            return self.func(options, *rest, **kwargs)

        return self.func(*args, **kwargs)

    def __sub__(self, other: Option):
        """Start an execution of this command by supplying options"""
        builder = CommandBuilder(self)
        return builder - other

    def _create_docstring(self):
        doc = [self.func.__doc__ or "Call {}".format(self.name)]
        if self.accepts:
            doc.append("FLAGS")
            for name, spec in sorted(self.accepts.items()):
                doc.append("\t-{}\t{}".format(name, spec.help))

        self.__call__.__func__.__doc__ = "\n".join(doc)


class CommandBuilder(CommandLike):
    """Used when a command is being built by options being passed to it"""

    def __init__(self, cmd: Command):
        self.command = cmd
        self.options = Options()

    def execute(self, *args, **kwargs):
        return self.command(self.options, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.execute(*args, **kwargs)

    def __sub__(self, other: Option):
        unsupported = ValueError("Command does not accept option: {}".format(other))

        if isinstance(other, Option):
            # We have an option instance so it could be an option with a value or the
            # arguments are actually meant for the command
            try:
                spec = self.command.accepts[other.name]
            except KeyError:
                raise unsupported
            else:
                params_stack = list(other.params_stack)

                if spec.is_flag:
                    self.options.add(other.name, True)
                    if params_stack:
                        # Assume the parameters are for the command, but make sure only one
                        assert len(params_stack) == 1, \
                            "Flag '{}' does not take any parameters".format(other.name)
                else:
                    option_params = params_stack.pop(0)  # tuple of (args, kwargs)
                    assert not option_params[1], "Can't supply kwargs to an option"
                    value = True
                    if option_params[0]:
                        value = option_params[0][0]
                    self.options.add(other.name, value)

                # Now the arguments of the option have been dealt with, see if we should execute
                if params_stack:
                    args, kwargs = params_stack.pop()
                    return self.execute(*args, **kwargs)

                # Carry on
                return self

        raise unsupported

    def __str__(self) -> str:
        return " ".join((self.command.name, str(self.options)))


# endregion


def option(opt: Option, help=''):  # pylint: disable=redefined-builtin
    """Decorator for defining an option taken by a command function"""
    spec = OptionSpec(opt, is_flag=False, help=help)

    def attach(func):
        if isinstance(func, Command):
            cmd = func
        else:
            cmd = Command(func)

        cmd.add_opt(spec)
        return cmd

    return attach


def flag(opt: Option, help=''):  # pylint: disable=redefined-builtin
    """Decorator for defining an option taken by a command function"""
    spec = OptionSpec(opt, is_flag=True, help=help)

    def attach(func):
        if isinstance(func, Command):
            cmd = func
        else:
            cmd = Command(func)

        cmd.add_opt(spec)
        return cmd

    return attach



__all__ = 'PyOSError', 'IsADirectoryError', 'NotADirectoryError', 'FileNotFoundError'

# pylint: disable=redefined-builtin


class PyOSError(Exception):
    """Raised when there is a PyOS exception"""


class IsADirectoryError(PyOSError):
    """Raise when a file is expected but a directory was passed"""


class NotADirectoryError(PyOSError):
    """Raise when a directory is expected but a file was passed"""


class FileNotFoundError(PyOSError):  # pylint: disable=redefined-builtin
    """A file was not found"""



"""Module with convenience functions for building queries"""
from typing import Iterable

import pyos


def or_(*conditions):
    if len(conditions) == 1:
        return conditions[0]

    return {'$or': list(conditions)}


def and_(*conditions):
    if len(conditions) == 1:
        return conditions

    return {'$and': list(conditions)}


def unset_(*keys: Iterable[str]):
    return {'$unset': {key: '' for key in keys}}


def in_(*args):
    if len(args) == 1:
        return args[0]

    return {'$in': list(args)}


def gt(val):  # pylint: disable=invalid-name
    return {'$gt': val}


gt_ = gt  # pylint: disable=invalid-name


def subdirs(root: str, start_depth=1, end_depth=1) -> dict:
    """Get a query string that will look in subdirectories of root optionally specifying the
    start and end depths
    """
    if start_depth in [0, 1] and end_depth == -1:
        match_root = '^{}'.format(root)
        and_below = '' if start_depth == 0 else '.+'
        regex = '{}{}'.format(match_root, and_below)
    else:
        if end_depth == -1:
            end_depth = ''  # This will cause the regex to allow any number of repetitions

        # The breakdown of this regexp is:
        # ^{root} - match strings beginning with the root
        # (
        #   [^/]+ - followed by a '/' or start of string one or more times
        #   /     - followed by exactly on occurrence of '/'
        # )
        # {{{start_depth},{end_depth}}} repeated a minimum of start depth and a maximum of
        # end_depth times
        regex = '^{}([^/]+/){{{},{}}}$'.format(root, start_depth, end_depth)

    return {pyos.config.DIR_KEY: {'$regex': regex}}


def dirmatch(directory: str) -> dict:
    """Get the query dictionary to search in a particular directory"""
    query = {pyos.config.DIR_KEY: str(directory)}
    if directory == "/":
        # Special case for root: all objects that have no DIR_KEY are by default
        # considered to be in the root
        query = or_(query, {pyos.config.DIR_KEY: {'$exists': False}})
    return query



import ast
import collections
import functools
from typing import Callable

from pytray import obj_load
from mincepy import qops

__all__ = ('type_string',)


def type_string(value: str) -> type:
    """Helper to get types as an argparse parameter"""
    return obj_load.load_obj(value)


def parse_query(value: str) -> dict:
    for oper, func in OPERATORS.items():
        if oper in value:
            left, right = value.split(oper, maxsplit=1)

            # Try getting python types from the right hand side (e.g. None, False, 3, 2.6, etc)
            try:
                right = ast.literal_eval(right.rstrip())
            except ValueError:
                pass

            return func(left.strip(), right)

    raise ValueError("Unknown condition: {}".format(value))


def apply(operator: Callable, name: str, value):
    return {name: operator(value)}


def _new(operator: Callable):
    return functools.partial(apply, operator)


OPERATORS = collections.OrderedDict([
    ('!=', _new(qops.ne_)),
    ('=', _new(lambda x: x)),
    ('>', _new(qops.gt_)),
    ('<', _new(qops.lt_)),
])



import abc
from collections.abc import Sequence, Set
import copy
import functools
import sys
import typing

import anytree
import columnize
import tabulate

import mincepy

from pyos import config
from pyos import db
from pyos import exceptions
from pyos import fmt
from pyos import os
from pyos import pathlib
from pyos import results
from pyos import utils

__all__ = ('BaseNode', 'DirectoryNode', 'ObjectNode', 'ResultsNode', 'to_node', 'TABLE_VIEW',
           'LIST_VIEW', 'TREE_VIEW')

LIST_VIEW = 'list'
TREE_VIEW = 'tree'
TABLE_VIEW = 'table'


class BaseNode(Sequence, anytree.NodeMixin, results.BaseResults, metaclass=abc.ABCMeta):

    def __init__(self, name: str, parent=None):
        super().__init__()
        self._name = name
        self.parent = parent
        self._hist = db.get_historian()

    def __getitem__(self, item):
        return self.children.__getitem__(item)

    def __len__(self):
        return self.children.__len__()

    @property
    def name(self):
        return self._name

    def delete(self):
        """Delete this node and any descendents"""
        for child in self.children:
            child.delete()

    def move(self, dest: os.PathSpec, overwrite=False):
        """Move this object (with any children) into the directory given by dest

        :param dest: the destination to move the node to
        :param overwrite: overwrite if exists
        """
        dest = pathlib.Path(dest).to_dir()
        for child in self.children:
            child.move(dest, overwrite)


class FilesystemNode(BaseNode):
    """Base node for representing an object in the virtual filesystem"""

    def __init__(self, path: os.PathSpec, parent: BaseNode = None):
        """
        :param path: the path this node represents
        :param parent: parent node
        """
        path = pathlib.Path(path).resolve()
        super().__init__(path.name, parent)
        self._abspath = path

    @property
    def abspath(self) -> 'pathlib.Path':
        return self._abspath

    @abc.abstractmethod
    def rename(self, new_name: str):
        """Rename this filesystem node"""


class ContainerNode(BaseNode):
    """A node that contains children that can be either directory nodes or object nodes"""

    def __contains__(self, item):
        # pylint: disable=too-many-nested-blocks, too-many-branches, too-many-return-statements
        if isinstance(item, pathlib.PurePath):
            path = pathlib.PurePath(item)
            if path.is_absolute():
                if path.is_dir_path():
                    # It's a directory
                    for node in self.directories:
                        if path == node.abspath:
                            return True
                else:
                    # It's a filename
                    for node in self.objects:
                        if path == node.abspath:
                            return True

                # Always check within directories
                for node in self.directories:
                    if path in node:
                        return True
            else:
                # It's relative
                parts = path.parts
                if len(parts) > 1:
                    subpath = pathlib.PurePath("".join(parts[1:]))

                    # Check subdirs
                    for node in self.directories:
                        if node.abspath.name == parts[0] and subpath in node:
                            return True
                else:
                    if path.is_dir_path():
                        # It's a directory
                        for node in self.directories:
                            if node.abspath.name == parts[0]:
                                return True
                    else:
                        # It's a filename
                        for obj in self.objects:
                            if obj.abspath.name == parts[0]:
                                return True

            return False

        if self._hist.is_obj_id(item):
            for node in self.objects:
                if item == node.obj_id:
                    return True

            return False

        return False

    def __getitem__(self, item):
        items = super(ContainerNode, self).__getitem__(item)
        if isinstance(item, slice):
            res = ResultsNode()
            for entry in items:
                res.append(copy.copy(entry))
            return res

        return items

    @property
    def directories(self):
        return filter(lambda node: isinstance(node, DirectoryNode), self.children)

    @property
    def objects(self):
        return filter(lambda node: isinstance(node, ObjectNode), self.children)


class DirectoryNode(ContainerNode, FilesystemNode):

    def __init__(self, path: os.PathSpec, parent: BaseNode = None):
        path = pathlib.PurePath(path)
        assert path.is_dir_path(), "Must supply a directory path"
        super().__init__(path.resolve(), parent)

    def __repr__(self):
        rep = []
        for pre, _, node in anytree.RenderTree(self):
            rep.append("%s%s" % (pre, node.name))
        return "\n".join(rep)

    def __copy__(self):
        """Create a copy with no parent"""
        dir_node = DirectoryNode(self.abspath)
        dir_node.children = [copy.copy(child) for child in self.children]
        return dir_node

    def _invalidate_cache(self):
        self.children = []

    def expand(self, depth=1, populate_objects=False):
        """Populate the children with what is currently in the database

        :param depth: expand to the given depth, 0 means no expansion, 1 means my child nodes, etc
        :param populate_objects: if True objects will have their records fetched immediately (as
            opposed to lazily when needed).  This gives a large speedup when the client knows that
            the all or most of the details of the child objects will be needed as they can be
            fetched in one call.
        """
        self.children = []
        if depth == 0:
            return

        path_str = str(self._abspath)

        # OBJECTS
        # Query for objects in this directory
        obj_kwargs = []
        for obj_id, meta in self._hist.meta.find({config.DIR_KEY: path_str}):
            # This is an object in _this_ directory and not further down the hierarchy
            obj_kwargs.append(dict(obj_id=obj_id, meta=meta, parent=self))

        # DIRECTORIES

        # Search for objects in directories below this one (depth 1) to any depth (-1) as
        # there may be just folders (no objects) between _abspath and the object in some deeply
        # nested directory
        directories = self._hist.meta.distinct(config.DIR_KEY,
                                               {config.DIR_KEY: {
                                                   '$regex': '^' + path_str
                                               }})

        len_path_str = len(path_str)

        child_expand_depth = depth - 1
        directories_added = set()  # type: Set[str]

        for dirstring in directories:
            if dirstring == path_str:
                # Skip over this directory
                continue

            # This is an object that resides at some subdirectory so we know
            # that there must exist a subdirectory
            dirname = dirstring[len_path_str:].split(os.sep)[0]

            if dirname not in directories_added:
                directories_added.add(dirname)
                # Get the subdirectory relative to us
                path = (self._abspath / dirname).to_dir()

                dir_node = DirectoryNode(path, parent=self)
                if abs(child_expand_depth) > 0:
                    dir_node.expand(child_expand_depth)

        if populate_objects:
            # Gather all the object ids
            obj_ids = [kwargs['obj_id'] for kwargs in obj_kwargs]
            records = {record.obj_id: record for record in self._hist.archive.find(obj_id=obj_ids)}
            for kwargs in obj_kwargs:
                kwargs['record'] = records[kwargs['obj_id']]

        for kwargs in obj_kwargs:
            ObjectNode(**kwargs)

    def delete(self):
        with self._hist.transaction():
            for obj_id, _meta in self._hist.meta.find(db.queries.subdirs(str(self._abspath), 0,
                                                                         -1)):
                self._hist.delete(obj_id)
        self._invalidate_cache()

    def move(self, dest: os.PathSpec, overwrite=False):
        dest = pathlib.Path(dest).to_dir().resolve() / self.name
        os.rename(self._abspath, dest)
        self._abspath = dest

    def rename(self, new_name: str):
        new_path = pathlib.Path(self.abspath.parent / new_name).to_dir()
        os.rename(self.abspath, new_path)
        self._abspath = new_path


class ObjectNode(FilesystemNode):
    """A node that represents an object"""

    @classmethod
    def from_path(cls, path: os.PathLike):
        path = pathlib.PurePath(path)
        if path.is_dir_path():
            raise exceptions.IsADirectoryError(path)

        path = path.resolve()
        meta = None
        # Let's find the object id
        obj_id = db.to_obj_id(path.name)
        if obj_id is None:
            # Ok, have to do a lookup
            res = tuple(db.find_meta(db.path_to_meta_dict(path)))
            if not res:
                raise ValueError("'{}' is not a valid object path".format(path))

            obj_id, meta = res[0]

        return ObjectNode(obj_id, meta=meta)

    def __init__(self, obj_id, record: mincepy.DataRecord = None, meta=None, parent=None):
        if record:
            assert obj_id == record.obj_id, "Obj id and record don't match!"

        self._hist = db.get_historian()
        self._obj_id = obj_id
        self._record = record  # This will be lazily loaded if None

        # Set up the meta
        self._meta = meta
        if self._meta is None:
            self._meta = self._hist.meta.get(obj_id)
            if self._meta is None:
                # Still don't have metadata! Double check this object actually exists
                try:
                    record = next(self._hist.archive.find(obj_id=obj_id,
                                                          deleted=True))  # type: mincepy.DataRecord
                except StopIteration:
                    raise ValueError("Object with id '{}' not found".format(obj_id))
                else:
                    if record.is_deleted_record():
                        raise ValueError("Object with id '{}' has been deleted".format(obj_id))
                    self._record = record
                    self._meta = {}

        # Set up the abspath
        abspath = db.get_abspath(obj_id, self._meta)
        super().__init__(abspath, parent)

    def __contains__(self, item):
        """Object nodes have no children and so do not contain anything"""
        return False

    def __copy__(self):
        """Make a copy with no parent"""
        return ObjectNode(self._obj_id, meta=self._meta, record=self._record)

    @property
    def record(self) -> mincepy.DataRecord:
        if self._record is None:
            # Lazily load
            self._record = tuple(self._hist.archive.find(
                self._obj_id))[0]  # type: mincepy.DataRecord

        return self._record

    @property
    def loaded(self):
        try:
            self._hist.get_obj(self._obj_id)
            return True
        except mincepy.NotFound:
            return False

    @property
    def obj(self) -> typing.Any:
        return self._hist.load(self._obj_id)

    @property
    def obj_id(self):
        return self._obj_id

    @property
    def type_id(self):
        return self.record.type_id

    @property
    def type(self):
        try:
            return self._hist.get_obj_type(self.record.type_id)
        except TypeError:
            return self.type_id

    @property
    def ctime(self):
        return self.record.creation_time

    @property
    def version(self):
        return self.record.version

    @property
    def mtime(self):
        return self.record.snapshot_time

    @property
    def creator(self):
        return self.record.get_extra(mincepy.ExtraKeys.CREATED_BY)

    @property
    def meta(self) -> dict:
        return self._meta

    def delete(self):
        self._hist.delete(self._obj_id)

    def move(self, dest: os.PathSpec, overwrite=False):
        dest = pathlib.Path(dest).to_dir().resolve() / self.name
        db.rename(self.obj_id, dest)
        self._abspath = dest

    def rename(self, new_name: str):
        new_name = self.abspath.parent / new_name
        if new_name.is_dir():
            raise exceptions.IsADirectoryError(new_name)

        try:
            db.rename(self._obj_id, new_name)
        except mincepy.DuplicateKeyError:
            raise RuntimeError("File with the name '{}' already exists".format(new_name))


class ResultsNode(ContainerNode):
    VIEW_PROPERTIES = ('loaded', 'type', 'creator', 'version', 'ctime', 'mtime', 'name', 'str')

    def __init__(self, name='results', parent=None):
        super().__init__(name, parent)
        self._view_mode = TABLE_VIEW
        self._show = {'name'}

    def __repr__(self):
        if self._view_mode == TREE_VIEW:
            rep = []
            for child in self.directories:
                for pre, _, node in anytree.RenderTree(child):
                    rep.append("%s%s" % (pre, node.name))
            for child in self.objects:
                rep.append(str(child))
            return "\n".join(rep)

        if self._view_mode == TABLE_VIEW:
            rep = []

            if self._deeply_nested():
                # Do the objects first, like linux's 'ls'
                table = self._get_table(self.objects)
                rep.append(tabulate.tabulate(table, tablefmt='plain'))

                for directory in self.directories:
                    dir_repr = []
                    dir_repr.append("{}:".format(directory.name))
                    table = self._get_table(directory)
                    dir_repr.append(tabulate.tabulate(table, tablefmt='plain'))
                    rep.append("\n".join(dir_repr))
            else:
                my_repr = []
                table = self._get_table(self.directories)
                table.extend(self._get_table(self.objects))
                my_repr.append(tabulate.tabulate(table, tablefmt='plain'))
                rep.append("\n".join(my_repr))

            return "\n\n".join(rep)

        if self._view_mode == LIST_VIEW:

            repr_list = []
            for child in self:
                repr_list.append("-".join(self._get_row(child)))

            if sys.stdout.isatty():
                return columnize.columnize(repr_list, displaywidth=utils.get_terminal_width())
            else:
                return "\n".join(repr_list)

        return super().__repr__()

    def __getitem__(self, item):
        result = super(ResultsNode, self).__getitem__(item)
        if isinstance(result, ResultsNode):
            # Transfer the view mode
            result.show(*self._show, mode=self._view_mode)
        return result

    @property
    def showing(self) -> set:
        """Returns the current view properties that are being displayed (if the view mode supports
        them)"""
        return self._show

    @property
    def view_mode(self) -> str:
        return self._view_mode

    @view_mode.setter
    def view_mode(self, new_mode: str):
        assert new_mode in (TREE_VIEW, LIST_VIEW, TABLE_VIEW)
        self._view_mode = new_mode

    def append(self, node: FilesystemNode, display_name: str = None):
        """Append a node to the results"""
        node.parent = self
        display_name = display_name or node.name
        node.display_name = display_name

    def extend(self, other: ContainerNode):
        """Extend this results using incorporating the entries of the other container"""
        for entry in other:
            self.append(entry)

    def show(self, *properties, mode: str = None):
        if mode is not None:
            self._view_mode = mode
        if properties:
            self._show = set(properties)

    def _get_row(self, child) -> typing.Sequence[str]:
        # pylint: disable=too-many-branches
        empty = ''
        row = []

        if 'loaded' in self._show:
            try:
                row.append('*' if child.loaded else '')
            except AttributeError:
                row.append(empty)

        if 'type' in self._show:
            try:
                row.append(fmt.pretty_type_string(child.type))
            except AttributeError:
                row.append('directory')

        if 'creator' in self._show:
            row.append(getattr(child, 'creator', empty))

        if 'version' in self._show:
            row.append(getattr(child, 'version', empty))

        if 'ctime' in self._show:
            try:
                row.append(fmt.pretty_datetime(child.ctime))
            except AttributeError:
                row.append(empty)

        if 'mtime' in self._show:
            try:
                row.append(fmt.pretty_datetime(child.mtime))
            except AttributeError:
                row.append(empty)

        if 'name' in self._show:
            row.append(getattr(child, 'name', empty))

        if 'str' in self._show:
            try:
                row.append(str(getattr(child, 'obj', empty))[:30])
            except (TypeError, mincepy.ObjectDeleted):
                row.append(empty)

        if 'abspath' in self._show:
            row.append(str(getattr(child, 'abspath', empty)))

        if 'relpath' in self._show:
            try:
                row.append(os.path.relpath(child.abspath))
            except AttributeError:
                row.append(empty)

        return row

    def _get_table(self, entry) -> list:
        table = []

        for child in entry:
            table.append(self._get_row(child))

        return table

    def _deeply_nested(self) -> bool:
        """Returns True if we have any nodes that themselves have children"""
        for directory in self.directories:
            if len(directory) > 0:
                return True

        return False


@functools.singledispatch
def to_node(entry) -> FilesystemNode:
    """Get the node for a given object.  This can be either:

    1.  An object id -> ObjectNode
    2.  A directory path -> DirectoryNode
    3.  An object path -> ObjectNode
    """
    if db.get_historian().is_obj_id(entry):
        return ObjectNode(entry)

    raise ValueError("Unknown entry type: {}".format(entry))


@to_node.register(FilesystemNode)
def _(entry: FilesystemNode):
    return entry


@to_node.register(os.PathLike)
def _(entry: os.PathLike):
    # Make sure we've got a pure path so we don't actually check that database
    entry = pathlib.PurePath(entry)

    if entry.is_dir_path():
        return DirectoryNode(entry)

    return ObjectNode.from_path(entry)



from typing import Iterable, List, Optional, Callable

import pyos
from pyos import glob
from pyos import os as pos
from . import shell

__all__ = ('PathCompletion',)


class PathCompletion(pyos.os.PathLike):
    """This class enables tab-completion for paths.  Passed a path, this class will list all the
    files and directories below it when __dir__ is called (i.e. tab in, e.g., ipython)."""

    def __init__(self, path='.'):
        self._path = pyos.pathlib.Path(path)

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self._path)

    def __fspath__(self):
        return str(self._path)

    def __dir__(self) -> Iterable[str]:
        try:
            for path in self._path.to_dir().iterdir():
                name = path.name
                if name.endswith('/'):
                    name = name[:-1]

                # Skip strings that aren't python identifiers
                if name.isidentifier():
                    yield name

        except (pyos.FileNotFoundError, pyos.NotADirectoryError):
            pass

    def _ipython_key_completions_(self):
        """This allows getitem style (["<tab>) style completion"""
        try:
            for path in self._path.to_dir().iterdir():
                yield path.name

        except (pyos.FileNotFoundError, pyos.NotADirectoryError):
            pass

    def __getattr__(self, item: str) -> 'PathCompletion':
        """Attribute access for paths"""
        # This creates a file (because forward slash isn't a valid python variable name character),
        # but it may be a directory..

        path = self._path / item
        # ..check:
        if not path.exists():
            path = path.to_dir()
            if not path.exists():
                try:
                    return super().__getattr__(item)
                except AttributeError:
                    raise AttributeError("Path does not exist: '{}'".format(item))

        return PathCompletion(path)

    def __getitem__(self, item):
        """Square bracket notation.  Allows arbitrary strings which is needed if the path name isn't
        a valid python variable"""
        path = self._path / item
        if path.exists():
            return PathCompletion(path)

        path = path.to_dir()
        if path.exists():
            return PathCompletion(path)

        raise FileNotFoundError("Path does not exist: '{}'".format(item))


def path_complete(app: shell.PyosShell,
                  text: str,
                  _line: str,
                  _begidx: int,
                  _endidx: int,
                  *,
                  path_filter: Optional[Callable[[str], bool]] = None) -> List[str]:
    """Performs completion of local file system paths

    :param app: The pyos base shell
    :param text: the string prefix we are attempting to match (all matches must begin with it)
    :param _line: the current input line with leading whitespace removed
    :param _begidx: the beginning index of the prefix text
    :param _endidx: the ending index of the prefix text
    :param path_filter: optional filter function that determines if a path belongs in the
        results this function takes a path as its argument and returns True if the path should
        be kept in the results
    :return: a list of possible tab completions
    """
    # If the search text is blank, then search in the CWD for *
    if not text:
        search_str = '*'
    else:
        # Purposely don't match any path containing wildcards
        if '*' in text or '?' in text:
            return []

        # Start the search string
        search_str = text + '*'

    # Set this to True for proper quoting of paths with spaces
    app.matches_delimited = True

    # Find all matching path completions
    matches = glob.glob(search_str)

    # Filter out results that don't belong
    if path_filter is not None:
        matches = [c for c in matches if path_filter(c)]

    # Don't append a space or closing quote to directory
    if len(matches) == 1 and pos.path.isdir(matches[0]):
        app.allow_appended_space = False
        app.allow_closing_quote = False

    # Sort the matches before any trailing slashes are added
    matches.sort(key=app.default_sort_key)
    app.matches_sorted = True

    return matches


def file_completer(app: shell.PyosShell, text: str, line: str, begidx: int,
                   endidx: int) -> List[str]:

    def is_file(path: str):
        return not path.endswith(pos.sep)

    return path_complete(app, text, line, begidx, endidx, path_filter=is_file)


def dir_completer(app: shell.PyosShell, text: str, line: str, begidx: int,
                  endidx: int) -> List[str]:

    def is_dir(path: str):
        return path.endswith(pos.sep)

    return path_complete(app, text, line, begidx, endidx, path_filter=is_dir)



"""The connection commands"""
import argparse

import cmd2
import mincepy

from pyos import db


class Connnection(cmd2.CommandSet):
    connect_parser = argparse.ArgumentParser()
    connect_parser.add_argument('uri', type=str, help='the URI of the archive to connect to')

    @cmd2.with_argparser(connect_parser)
    def do_connect(self, app: cmd2.Cmd, args):  # pylint: disable=no-self-use
        """Connect to a mincepy archive"""
        try:
            db.connect(args.uri)
            app.poutput("Connected to {}".format(args.uri))
        except mincepy.ConnectionError as exc:
            app.perror("Error: {}".format(exc))

    def do_disconnect(self, _app: cmd2.Cmd, _cmd):  # pylint: disable=no-self-use
        """Disconnect from the current archive"""
        db.reset()



# -*- coding: utf-8 -*-
from setuptools import setup

__author__ = "Martin Uhrin"
__license__ = "GPLv3"
__contributors__ = "Sonia Collaud"

about = {}
with open('pyos/version.py') as f:
    exec(f.read(), about)

setup(
    name='pyos',
    version=about['__version__'],
    description="Browse python objects as if they were files on disk",
    long_description=open('README.rst').read(),
    url='https://github.com/muhrin/pyos.git',
    author='Martin Uhrin',
    author_email='martin.uhrin.10@ucl.ac.uk',
    license=__license__,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='database schemaless nosql object-store',
    install_requires=[
        'anytree',
        'cmd2 ~= 1.3.2',
        'columnize',
        'mincepy>=0.15.4, <0.16.0',
        'click',
        'tabulate',
        'ipython',
        'pyprnt',
        'pytray >= 0.3.0',
        'stevedore',
        'tqdm',
    ],
    extras_require={
        'gui': ['mincepy[gui]'],
        'dev': [
            'cmd2-ext-test',
            'ipython',
            'jupyter-sphinx',
            'nbsphinx',
            'pip',
            'pytest>4',
            'pytest-cov',
            'pre-commit',
            'sphinx',
            'sphinx-autobuild',
            'yapf',
            'prospector',
            'pylint',
            'twine',
        ],
    },
    packages=['pyos', 'pyos.db', 'pyos.fs', 'pyos.os', 'pyos.psh', 'pyos.psh.cmds', 'pyos.psh_lib'],
    project_urls={
        'Documentation': 'https://pyos.readthedocs.org/',
        'Source': 'https://github.com/muhrin/pyos/',
    },
    include_package_data=True,
    test_suite='test',
    entry_points={
        'console_scripts': ['pyos = pyos.cli:pyos', 'psh = pyos.cli:psh_'],
        'mincepy.plugins.types': ['pyos_types = pyos.provides:get_types',]
    })



"""Module that deals with directories and paths"""
import contextlib
from typing import Sequence, Iterable, Tuple, Optional
import uuid

import mincepy

from pyos import db
from pyos import exceptions
from pyos import fs
from pyos import os

__all__ = ('Path', 'PurePath')


class PurePath(os.PathLike):
    """A path in PyOS.  Where possible the convention follows that of a PurePosixPath in pathlib.
    The one major exception is that folders are represented with an explicit trailing '/' and
    anything else is a file.

    This is a 'pure' path in a similar sense to pathlib.PurePath in that it does not interact with
    the database at all.
    """

    def __init__(self, path='./'):
        super().__init__()
        self._path = os.path.normpath(path)

    @property
    def parts(self) -> Tuple[str]:
        parts = list(self._path.split('/'))
        for idx in range(len(parts) - 1):
            parts[idx] += '/'
        if self.is_dir_path():
            parts.pop()

        return tuple(parts)

    @property
    def parent(self):
        parts = self.parts

        if self.is_absolute():
            if len(parts) == 1:
                return self.__class__(self.root)

            return self.__class__(''.join(parts[:-1]))

        # Relative path
        if len(parts) == 1:
            return self.__class__()

        return self.__class__(''.join(parts[:-1]))

    @property
    def parents(self) -> Sequence:
        """Get a sequence of paths that are the parents of this path"""
        current = self.parent
        parents = [current]
        while current.parent != parents[-1]:
            parents.append(current.parent)
            current = parents[-1]
        return parents

    @property
    def name(self):
        return self.parts[-1]

    @property
    def root(self) -> str:
        return "/"

    def __fspath__(self):
        return self._path

    def __hash__(self):
        return self._path.__hash__()

    def __eq__(self, other):
        if not isinstance(other, PurePath):
            return False

        return self._path == other._path

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self._path)

    def __str__(self):
        return self._path

    def __truediv__(self, other):
        if not isinstance(other, PurePath):
            if isinstance(other, str):
                other = self.__class__(other)
            else:
                raise TypeError("Cannot join a path with a '{}'".format(type(other)))

        if self.is_file_path():
            raise ValueError("Can't join a file with another path")

        if other.is_absolute():
            return other

        return self.__class__(str(self) + str(other))

    def is_file_path(self) -> bool:
        """Returns True if this path specifies a file i.e. does not end with a training '/'"""
        return not self.is_dir_path()

    def is_dir_path(self) -> bool:
        """Returns True if this path specified a directory i.e. ends with a trailing '/'"""
        return self._path.endswith(os.sep)

    def is_absolute(self) -> bool:
        return os.path.isabs(self._path)

    def to_dir(self) -> 'PurePath':
        """If this path is a file path then a directory with the same name will be
        returned. Otherwise this path will be returned"""
        if self.is_dir_path():
            return self

        return self.__class__(str(self) + '/')

    def to_file(self):
        """If this path is a directory then a file path with the same name will be returned.
        Otherwise this path will be returned"""
        if self.is_file_path():
            return self

        return self.__class__(str(self)[:-1])

    def resolve(self):
        """Make the path absolute eliminating any . and .. that occur in the path"""
        path = os.path.abspath(self)
        return self.__class__(path)

    def joinpath(self, *other):
        return self.__class__(os.path.join(self, *other))


class Path(PurePath, mincepy.SimpleSavable):
    """A path in Pyos.  Where possible the convention follows that of a PurePosixPath in pathlib.
    The one major exception is that folders are represented with an explicit trailing '/' and
    anything else is a file."""
    ATTRS = ('_path',)
    TYPE_ID = uuid.UUID('5eac541e-848c-43aa-818d-50cf8a2b8507')

    def is_file(self) -> bool:
        """Returns True if this path is a file path and exists"""
        return self.is_file_path() and self.exists()

    def is_dir(self) -> bool:
        """Returns True if this path is a directory path and exists"""
        return self.is_dir_path() and self.exists()

    def exists(self) -> bool:
        """Test whether a path point exists"""
        return os.path.exists(self)

    def unlink(self, missing_ok=False):
        if not self.exists():
            if missing_ok:
                return
            raise exceptions.FileNotFoundError("Can't delete '{}', it does not exist".format(self))

        os.unlink(self)

    def iterdir(self) -> Iterable['Path']:
        """
        When the path points to a directory, yield path objects of the directory contents:

        >>>
        >>> p = Path('docs')
        >>> for child in p.iterdir(): child
        ...
        Path('docs/conf')
        Path('docs/readme')
        Path('docs/index.rst')
        Path('docs/_build')
        Path('docs/_static')
        """
        if not self.is_dir_path():
            raise exceptions.NotADirectoryError("Not a directory: {}".format(os.path.relpath(self)))
        if not self.exists():
            raise exceptions.FileNotFoundError("No such directory: '{}'".format(
                os.path.relpath(self)))

        node = fs.DirectoryNode(self)
        node.expand(1)
        for child in node.children:
            yield child.abspath

    def rename(self, target: os.PathSpec) -> 'Path':
        target = Path(target).resolve()
        os.rename(self._path, target)
        return target


@contextlib.contextmanager
def working_path(path: os.PathSpec):
    """Context manager that changes the current working directory for the duration of the context,
    returning to the previous directory on exiting"""
    orig = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(orig)


def get_path(obj_or_id) -> Optional[Path]:
    """Given an object or object id get the current path"""
    return Path(db.get_path(obj_or_id))


def get_paths(*obj_or_id) -> Sequence[Path]:
    """Given objects or identifier this will return their current paths in the order they were
    passed"""
    return list(map(Path, db.get_paths(*obj_or_id)))



"""The history command"""

import pyos
from pyos import db
from .cat import cat


@pyos.psh_lib.command()
def history(obj):
    hist = db.get_historian()
    for entry in hist.history(obj):
        cat(entry.obj)



"""Module for built-in commands"""

# Commands:
from .cat import cat
from .cd import cd
from .find import find
from .history import history
from .load import load
from .locate import locate
from .log import Log
from .ls import ls
from .meta import meta
from .mv import mv
from .oid import oid
from .pwd import pwd
from .rm import rm
from .save import save
from .tree import tree

from . import connect

__all__ = ('cat', 'cd', 'find', 'history', 'load', 'ls', 'locate', 'meta', 'mv', 'oid', 'pwd', 'rm',
           'save', 'tree', 'Log')



"""The object id command"""

import mincepy

import pyos
from pyos import db


@pyos.psh_lib.command()
def oid(*args):
    """Get the object id for one or more live objects"""
    if not args:
        return None

    hist = db.get_historian()

    oids = []
    for obj in args:
        try:
            oids.append(hist.get_obj_id(obj))
        except mincepy.NotFound:
            oids.append(None)

    if len(oids) == 1:
        return oids[0]

    return oids



"""Classes and functions related to pyos' virtual filesystem.  This module is modelled on python's
os module and users familiar that will find many of the methods familiar."""

from .types import *
from .pos import *
from . import pos
from . import path

_ADDITIONAL = ('path',)

__all__ = types.__all__ + pos.__all__ + _ADDITIONAL



import mincepy

from . import db

__all__ = 'init', 'reset'


def connect(uri: str = '') -> mincepy.Historian:
    return db.connect(uri)


def init() -> mincepy.Historian:
    return db.init()


def reset():
    db.reset()
